import itertools
import json
import logging
import math
import os
import re
import uuid
from typing import (
    Any,
    cast,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TYPE_CHECKING,
)

from packaging.version import Version

from galaxy.tool_util.deps import requirements
from galaxy.tool_util.parser.util import (
    DEFAULT_DECOMPRESS,
    DEFAULT_DELTA,
    DEFAULT_DELTA_FRAC,
    DEFAULT_EPS,
    DEFAULT_METRIC,
    DEFAULT_PIN_LABELS,
    DEFAULT_SORT,
)
from galaxy.tool_util_models.parameter_validators import AnyValidatorModel
from galaxy.tool_util_models.tool_source import (
    Citation,
    DrillDownOptionsDict,
    FileSourceConfigFile,
    HelpContent,
    InputConfigFile,
    OutputCompareType,
    TemplateConfigFile,
    XmlTemplateConfigFile,
    XrefDict,
)
from galaxy.util import (
    Element,
    ElementTree,
    string_as_bool,
    string_as_bool_or_none,
    unicodify,
    XML,
    xml_text,
    xml_to_string,
)
from .interface import (
    AssertionList,
    DrillDownDynamicOptions,
    DynamicOptions,
    InputSource,
    PageSource,
    PagesSource,
    RequiredFiles,
    TestCollectionDefElementDict,
    TestCollectionDefElementObject,
    TestCollectionOutputDef,
    ToolSource,
    ToolSourceTest,
    ToolSourceTestInput,
    ToolSourceTestInputAttributes,
    ToolSourceTestInputs,
    ToolSourceTestOutput,
    ToolSourceTestOutputAttributes,
    ToolSourceTestOutputs,
    ToolSourceTests,
    XmlTestCollectionDefDict,
)
from .output_actions import (
    ToolOutputActionApp,
    ToolOutputActionGroup,
)
from .output_collection_def import dataset_collector_descriptions_from_elem
from .output_objects import (
    ChangeFormatModel,
    ToolExpressionOutput,
    ToolOutput,
    ToolOutputCollection,
    ToolOutputCollectionStructure,
)
from .parameter_validators import parse_xml_validators
from .stdio import (
    aggressive_error_checks,
    error_on_exit_code,
    StdioErrorLevel,
    ToolStdioExitCode,
    ToolStdioRegex,
)

if TYPE_CHECKING:
    from galaxy.util.path import StrPath
    from .output_objects import ToolOutputBase

log = logging.getLogger(__name__)


def inject_validates(inject):
    if inject == "api_key":
        return True
    elif inject == "entry_point_path_for_label":
        return True
    p = re.compile("^oidc_(id|access|refresh)_token_(.*)$")
    match = p.match(inject)
    return match is not None


def destroy_tree(tree):
    root = tree.getroot()

    node_tracker = {root: [0, None]}

    for node in root.iterdescendants():
        parent = node.getparent()
        node_tracker[node] = [node_tracker[parent][0] + 1, parent]

    node_tracker = sorted(
        [(depth, parent, child) for child, (depth, parent) in node_tracker.items()], key=lambda x: x[0], reverse=True
    )

    for _, parent, child in node_tracker:
        if parent is None:
            break
        parent.remove(child)

    del tree


def parse_change_format(change_format: Iterable[Element]) -> List[ChangeFormatModel]:
    change_models: List[ChangeFormatModel] = []
    for change_elem in change_format:
        for when_elem in change_elem.findall("when"):
            value: Optional[str] = when_elem.get("value", None)
            format_: Optional[str] = when_elem.get("format", None)
            check: Optional[str] = when_elem.get("input", None)
            input_dataset: Optional[str] = None
            check_attribute: Optional[str] = None
            if check is not None:
                if "$" not in check:
                    check = f"${check}"
            else:
                input_dataset = when_elem.get("input_dataset", None)
                check_attribute = when_elem.get("attribute", None)
            change_models.append(
                ChangeFormatModel(
                    value=value,
                    format=format_,
                    input=check,
                    input_dataset=input_dataset,
                    check_attribute=check_attribute,
                )
            )
    return change_models


class XmlToolSource(ToolSource):
    """Responsible for parsing a tool from classic Galaxy representation."""

    language = "xml"

    def __init__(
        self, xml_tree: ElementTree, source_path: Optional["StrPath"] = None, macro_paths: Optional[List[str]] = None
    ) -> None:
        self.xml_tree = xml_tree
        self.root = self.xml_tree.getroot()
        self._source_path = source_path
        self._macro_paths = macro_paths or []
        self.legacy_defaults = Version(self.parse_profile()) == Version("16.01")
        self._string = xml_to_string(self.root)

    def to_string(self):
        return self._string

    def mem_optimize(self):
        destroy_tree(self.xml_tree)
        self.root = None
        self._xml_tree = None

    def parse_version(self) -> Optional[str]:
        return self.root.get("version", None)

    def parse_id(self):
        return self.root.get("id")

    def parse_tool_module(self):
        root = self.root
        if root.find("type") is not None:
            type_elem = root.find("type")
            module = type_elem.get("module", "galaxy.tools")
            cls = type_elem.get("class")
            return module, cls

        return None

    def parse_action_module(self):
        root = self.root
        action_elem = root.find("action")
        if action_elem is not None:
            module = action_elem.get("module")
            cls = action_elem.get("class")
            return module, cls
        else:
            return None

    def parse_tool_type(self):
        root = self.root
        return root.get("tool_type")

    def parse_name(self):
        return self.root.get("name") or self.parse_id()

    def parse_edam_operations(self):
        edam_ops = self.root.find("edam_operations")
        if edam_ops is None:
            return []
        return [edam_op.text for edam_op in edam_ops.findall("edam_operation")]

    def parse_edam_topics(self):
        edam_topics = self.root.find("edam_topics")
        if edam_topics is None:
            return []
        return [edam_topic.text for edam_topic in edam_topics.findall("edam_topic")]

    def parse_xrefs(self) -> List[XrefDict]:
        xrefs = self.root.find("xrefs")
        if xrefs is None:
            return []
        return [
            XrefDict(value=xref.text.strip(), type=str(xref.attrib["type"]))
            for xref in xrefs.findall("xref")
            if xref.get("type") and xref.text
        ]

    def parse_description(self) -> str:
        return xml_text(self.root, "description")

    def parse_icon(self) -> Optional[str]:
        icon_elem = self.root.find("icon")
        return icon_elem.get("src") if icon_elem is not None else None

    def parse_display_interface(self, default):
        return self._get_attribute_as_bool("display_interface", default)

    def parse_require_login(self, default):
        return self._get_attribute_as_bool("require_login", default)

    def parse_request_param_translation_elem(self):
        return self.root.find("request_param_translation")

    def parse_command(self):
        command_el = self._command_el
        return ((command_el is not None) and command_el.text) or None

    def parse_expression(self):
        """Return string containing command to run."""
        expression_el = self.root.find("expression")
        if expression_el is not None:
            expression_type = expression_el.get("type")
            if expression_type != "ecma5.1":
                raise Exception(f"Unknown expression type [{expression_type}] encountered")
            return expression_el.text
        return None

    def parse_environment_variables(self):
        environment_variables_el = self.root.find("environment_variables")
        if environment_variables_el is None:
            return []

        environment_variables = []
        for environment_variable_el in environment_variables_el.findall("environment_variable"):
            template = environment_variable_el.text
            inject = environment_variable_el.get("inject")
            if inject:
                assert inject_validates(inject)
            if inject == "entry_point_path_for_label":
                assert (
                    template
                ), 'Environment variable value must contain entry point label when inject="entry_point_path_for_label".'
            else:
                assert not (template and inject), "Cannot specify inject and environment variable template."
            definition = {
                "name": environment_variable_el.get("name"),
                "template": template,
                "inject": inject,
                "strip": string_as_bool(environment_variable_el.get("strip", False)),
            }
            environment_variables.append(definition)
        return environment_variables

    def parse_home_target(self):
        target = "job_home" if Version(self.parse_profile()) >= Version("18.01") else "shared_home"
        command_el = self._command_el
        command_legacy = (command_el is not None) and command_el.get("use_shared_home", None)
        if command_legacy is not None:
            target = "shared_home" if string_as_bool(command_legacy) else "job_home"
        return target

    def parse_tmp_target(self):
        # Default to not touching TMPDIR et. al. but if job_tmp is set
        # in job_conf then do. This is a very conservative approach that shouldn't
        # break or modify any configurations by default.
        return "job_tmp_if_explicit"

    def parse_interpreter(self):
        interpreter = None
        command_el = self._command_el
        if command_el is not None:
            interpreter = command_el.get("interpreter", None)
        if interpreter and not self.legacy_defaults:
            log.warning("Deprecated interpreter attribute on command element is now ignored.")
            interpreter = None
        return interpreter

    def parse_version_command(self):
        version_cmd = self.root.find("version_command")
        if version_cmd is not None:
            return version_cmd.text
        else:
            return None

    def parse_version_command_interpreter(self):
        if self.parse_version_command() is not None:
            version_cmd = self.root.find("version_command")
            version_cmd_interpreter = version_cmd.get("interpreter", None)
            if version_cmd_interpreter:
                return version_cmd_interpreter
        return None

    def parse_parallelism(self):
        parallelism = self.root.find("parallelism")
        parallelism_info = None
        if parallelism is not None and parallelism.get("method"):
            return ParallelismInfo(parallelism)
        return parallelism_info

    def parse_interactivetool(self):
        interactivetool_el = self.root.find("entry_points")
        rtt = []
        if interactivetool_el is None:
            return rtt
        for ep_el in interactivetool_el.findall("entry_point"):
            port = ep_el.find("port")
            assert port is not None, ValueError("A port is required for InteractiveTools")
            port = port.text.strip()
            url = ep_el.find("url")
            if url is not None:
                url = url.text.strip()
            name = ep_el.get("name", None)
            if name:
                name = name.strip()
            label = ep_el.get("label", None)
            if label:
                label = label.strip()
            requires_domain = string_as_bool(ep_el.attrib.get("requires_domain", False))
            requires_path_in_url = string_as_bool(ep_el.attrib.get("requires_path_in_url", False))
            requires_path_in_header_named = ep_el.get("requires_path_in_header_named", None)
            if requires_path_in_header_named:
                requires_path_in_header_named = requires_path_in_header_named.strip()
            rtt.append(
                dict(
                    port=port,
                    url=url,
                    name=name,
                    label=label,
                    requires_domain=requires_domain,
                    requires_path_in_url=requires_path_in_url,
                    requires_path_in_header_named=requires_path_in_header_named,
                )
            )
        return rtt

    def parse_hidden(self):
        hidden = xml_text(self.root, "hidden")
        if hidden:
            hidden = string_as_bool(hidden)
        return hidden

    def parse_redirect_url_params_elem(self):
        return self.root.find("redirect_url_params")

    def parse_sanitize(self):
        return self._get_option_value("sanitize", True)

    def parse_refresh(self):
        return self._get_option_value("refresh", False)

    def _get_option_value(self, key, default):
        root = self.root
        for option_elem in root.findall("options"):
            if key in option_elem.attrib:
                return string_as_bool(option_elem.get(key))
        return default

    @property
    def _command_el(self):
        return self.root.find("command")

    @property
    def _outputs_el(self):
        return self.root.find("outputs")

    def _get_attribute_as_bool(self, attribute, default, elem=None):
        if elem is None:
            elem = self.root
        return string_as_bool(elem.get(attribute, default))

    def parse_required_files(self) -> Optional[RequiredFiles]:
        required_files = self.root.find("required_files")
        if required_files is None:
            return None

        def parse_include_exclude_list(tag_name):
            as_list = []
            for ref in required_files.findall(tag_name):
                path = ref.get("path")
                assert path is not None, f'"path" must be specified in {tag_name}'
                path_type = ref.get("type", "literal")
                as_list.append({"path": path, "path_type": path_type})
            return as_list

        as_dict = {}
        as_dict["extend_default_excludes"] = self._get_attribute_as_bool(
            "extend_default_excludes", True, elem=required_files
        )
        as_dict["includes"] = parse_include_exclude_list("include")
        as_dict["excludes"] = parse_include_exclude_list("exclude")
        return RequiredFiles.from_dict(as_dict)

    def parse_requirements(self):
        return requirements.parse_requirements_from_xml(self.root, parse_resources_and_credentials=True)

    def parse_input_pages(self) -> "XmlPagesSource":
        return XmlPagesSource(self.root)

    def parse_provided_metadata_style(self):
        style = None
        out_elem = self._outputs_el
        if out_elem is not None and "provided_metadata_style" in out_elem.attrib:
            style = out_elem.attrib["provided_metadata_style"]

        if style is None:
            style = "legacy" if Version(self.parse_profile()) < Version("17.09") else "default"

        assert style in ["legacy", "default"]
        return style

    def parse_provided_metadata_file(self):
        provided_metadata_file = "galaxy.json"
        out_elem = self.root.find("outputs")
        if out_elem is not None and "provided_metadata_file" in out_elem.attrib:
            provided_metadata_file = out_elem.attrib["provided_metadata_file"]

        return provided_metadata_file

    def parse_outputs(self, app: Optional[ToolOutputActionApp] = None):
        out_elem = self.root.find("outputs")
        outputs: Dict[str, ToolOutputBase] = {}
        output_collections: Dict[str, ToolOutputCollection] = {}
        if out_elem is None:
            return outputs, output_collections

        data_dict: Dict[str, ToolOutput] = {}
        expression_dict: Dict[str, ToolExpressionOutput] = {}

        def _parse(data_elem, **kwds):
            output_def = self._parse_output(data_elem, app, **kwds)
            data_dict[output_def.name] = output_def
            return output_def

        for _ in out_elem.findall("data"):
            _parse(_)

        def _parse_expression(output_elem, app: Optional[ToolOutputActionApp] = None, **kwds):
            output_def = self._parse_expression_output(output_elem, app, **kwds)
            output_def.filters = output_elem.findall("filter")
            expression_dict[output_def.name] = output_def
            return output_def

        def _parse_collection(collection_elem: Element):
            name = collection_elem.get("name")
            assert name
            label = xml_text(collection_elem, "label")
            default_format = collection_elem.get("format", "data")
            collection_type = collection_elem.get("type", None)
            collection_type_source = collection_elem.get("type_source", None)
            collection_type_from_rules = collection_elem.get("type_from_rules", None)
            structured_like = collection_elem.get("structured_like", None)
            inherit_format = False
            inherit_metadata = False
            if structured_like:
                inherit_format = string_as_bool(collection_elem.get("inherit_format", None))
                inherit_metadata = string_as_bool(collection_elem.get("inherit_metadata", None))
            default_format_source = collection_elem.get("format_source", None)
            default_metadata_source = collection_elem.get("metadata_source", None)
            filters = collection_elem.findall("filter")

            dataset_collector_descriptions = None
            if collection_elem.find("discover_datasets") is not None:
                dataset_collector_descriptions = dataset_collector_descriptions_from_elem(collection_elem, legacy=False)
            structure = ToolOutputCollectionStructure(
                collection_type=collection_type,
                collection_type_source=collection_type_source,
                collection_type_from_rules=collection_type_from_rules,
                structured_like=structured_like,
                dataset_collector_descriptions=dataset_collector_descriptions,
            )
            output_collection = ToolOutputCollection(
                name,
                structure,
                label=label,
                filters=filters,
                default_format=default_format,
                inherit_format=inherit_format,
                inherit_metadata=inherit_metadata,
                default_format_source=default_format_source,
                default_metadata_source=default_metadata_source,
            )
            outputs[output_collection.name] = output_collection

            for data_elem in collection_elem.findall("data"):
                _parse(
                    data_elem,
                    default_format=default_format,
                    default_format_source=default_format_source,
                    default_metadata_source=default_metadata_source,
                )

            for data_elem in collection_elem.findall("data"):
                output_name = data_elem.get("name")
                assert output_name
                data = data_dict[output_name]
                assert data
                del data_dict[output_name]
                output_collection.outputs[output_name] = data
            output_collections[name] = output_collection

        for out_child in out_elem:
            if out_child.tag == "data":
                _parse(out_child)
            elif out_child.tag == "collection":
                _parse_collection(out_child)
            elif out_child.tag == "output":
                output_type = out_child.get("type")
                if output_type == "data":
                    _parse(out_child)
                elif output_type == "collection":
                    out_child.attrib["type"] = unicodify(out_child.get("collection_type"))
                    out_child.attrib["type_source"] = unicodify(out_child.get("collection_type_source"))
                    _parse_collection(out_child)
                else:
                    _parse_expression(out_child, app)
            else:
                log.warning(f"Unknown output tag encountered [{out_child.tag}]")

        for output_def in itertools.chain(data_dict.values(), expression_dict.values()):
            outputs[output_def.name] = output_def
        return outputs, output_collections

    def _parse_output(
        self,
        data_elem,
        app: Optional[ToolOutputActionApp] = None,
        default_format="data",
        default_format_source=None,
        default_metadata_source=None,
        expression_type=None,
    ):
        from_expression = data_elem.get("from")
        output = ToolOutput(data_elem.get("name"), from_expression=from_expression)
        output_format = data_elem.get("format", default_format)
        auto_format = string_as_bool(data_elem.get("auto_format", "false"))
        if auto_format and output_format != "data":
            raise ValueError("Setting format and auto_format is not supported at this time.")
        elif auto_format:
            output_format = "_sniff_"
        output.format = output_format
        output.change_format = parse_change_format(data_elem.findall("change_format"))
        output.format_source = data_elem.get("format_source", default_format_source)
        output.default_identifier_source = data_elem.get("default_identifier_source", "None")
        output.metadata_source = data_elem.get("metadata_source", default_metadata_source)
        output.parent = data_elem.get("parent", None)
        output.label = xml_text(data_elem, "label", None)
        output.count = int(data_elem.get("count", 1))
        output.filters = data_elem.findall("filter")
        output.from_work_dir = data_elem.get("from_work_dir", None)
        output.precreate_directory = data_elem.get("precreate_directory") or False
        profile_version = Version(self.parse_profile())
        if output.from_work_dir and profile_version < Version("21.09"):
            # We started quoting from_work_dir outputs in 21.09.
            # Prior to quoting, trailing spaces had no effect.
            # This ensures that old tools continue to work.
            output.from_work_dir = output.from_work_dir.strip()
        output.hidden = string_as_bool(data_elem.get("hidden", ""))
        if app is not None:
            # poor design here driven entirely by pragmatism in refactoring, ToolOutputActionGroup
            # belongs in galaxy-tool because it uses app heavily. Breaking the output objects
            # into app-aware things and dumb models would be a large project but superior design
            # and decomposition.
            output.actions = ToolOutputActionGroup(app, data_elem.find("actions"))
        output.dataset_collector_descriptions = dataset_collector_descriptions_from_elem(
            data_elem, legacy=self.legacy_defaults
        )
        return output

    def _parse_expression_output(self, output_elem, app: Optional[ToolOutputActionApp] = None, **kwds):
        output_type = output_elem.get("type")
        from_expression = output_elem.get("from")
        output = ToolExpressionOutput(
            output_elem.get("name"),
            output_type,
            from_expression,
        )
        output.path = output_elem.get("value")
        output.label = xml_text(output_elem, "label", None)

        output.hidden = string_as_bool(output_elem.get("hidden", ""))
        if app is not None:
            output.actions = ToolOutputActionGroup(app, output_elem.find("actions"))
        return output

    def parse_stdio(self):
        """
        parse error handling from command and stdio tag

        returns list of exit codes, list of regexes

        - exit_codes contain all non-zero exit codes (:-1 and 1:) if
          detect_errors is default (if not legacy), exit_code, or aggressive
        - the oom_exit_code if given and detect_errors is exit_code
        - exit codes and regexes from the stdio tag
          these are prepended to the list, i.e. are evaluated prior to regexes
          and exit codes derived from the properties of the command tag.
          thus more specific regexes of the same or more severe error level
          are triggered first.

        """

        command_el = self._command_el
        detect_errors = None
        if command_el is not None:
            detect_errors = command_el.get("detect_errors")

        if detect_errors and detect_errors != "default":
            if detect_errors == "exit_code":
                oom_exit_code = None
                if command_el is not None:
                    oom_exit_code = command_el.get("oom_exit_code", None)
                if oom_exit_code is not None:
                    int(oom_exit_code)
                exit_codes, regexes = error_on_exit_code(out_of_memory_exit_code=oom_exit_code)
            elif detect_errors == "aggressive":
                exit_codes, regexes = aggressive_error_checks()
            else:
                raise ValueError(f"Unknown detect_errors value encountered [{detect_errors}]")
        elif len(self.root.findall("stdio")) == 0 and not self.legacy_defaults:
            exit_codes, regexes = error_on_exit_code()
        else:
            exit_codes = []
            regexes = []

        if len(self.root.findall("stdio")) > 0:
            parser = StdioParser(self.root)
            exit_codes = parser.stdio_exit_codes + exit_codes
            regexes = parser.stdio_regexes + regexes

        return exit_codes, regexes

    def parse_strict_shell(self):
        command_el = self._command_el
        if Version(self.parse_profile()) < Version("20.09"):
            default = "False"
        else:
            default = "True"
        if command_el is not None:
            return string_as_bool(command_el.get("strict", default))
        else:
            return string_as_bool(default)

    def parse_help(self) -> Optional[HelpContent]:
        help_elem = self.root.find("help")
        if help_elem is None:
            return None

        help_format = help_elem.get("format", "restructuredtext")
        content = help_elem.text or ""
        return HelpContent(format=help_format, content=content)

    @property
    def macro_paths(self) -> List[str]:
        return self._macro_paths

    @property
    def source_path(self):
        return self._source_path

    def parse_tests_to_dict(self) -> ToolSourceTests:
        tests_elem = self.root.find("tests")
        tests: List[ToolSourceTest] = []
        rval: ToolSourceTests = dict(tests=tests)

        if tests_elem is not None:
            for i, test_elem in enumerate(tests_elem.findall("test")):
                profile = self.parse_profile()
                tests.append(_test_elem_to_dict(test_elem, i, profile))

        return rval

    def parse_profile(self) -> str:
        # Pre-16.04 or default XML defaults
        # - Use standard error for error detection.
        # - Don't run shells with -e
        # - Auto-check for implicit multiple outputs.
        # - Auto-check for $param_file.
        # - Enable buggy interpreter attribute.
        return self.root.get("profile", "16.01")

    def parse_license(self) -> Optional[str]:
        return self.root.get("license")

    def parse_citations(self) -> List[Citation]:
        """Return a list of citations."""
        citations: List[Citation] = []
        root = self.root
        citations_elem = root.find("citations")
        if citations_elem is None:
            return citations

        for citation_elem in citations_elem:
            try:
                citation = parse_citation_elem(citation_elem)
            except Exception:
                if Version(self.parse_profile()) < Version("24.2"):
                    continue
                else:
                    raise
            if citation:
                citations.append(citation)

        return citations

    def parse_python_template_version(self):
        python_template_version = self.root.get("python_template_version")
        if python_template_version is not None:
            python_template_version = Version(python_template_version)
        return python_template_version

    def parse_template_configfiles(self) -> Sequence[TemplateConfigFile]:
        configfiles: List[XmlTemplateConfigFile] = []
        if (conf_parent_elem := self.root.find("configfiles")) is not None:
            for conf_elem in conf_parent_elem.findall("configfile"):
                name = conf_elem.get("name")
                filename = conf_elem.get("filename", None)
                content = conf_elem.text
                configfiles.append(XmlTemplateConfigFile(name=name, filename=filename, content=content))
        return configfiles

    def parse_input_configfiles(self) -> Sequence[InputConfigFile]:
        config_files: list[InputConfigFile] = []
        if (conf_parent_elem := self.root.find("configfiles")) is not None:
            inputs_elem = conf_parent_elem.find("inputs")
            if inputs_elem is not None:
                name = inputs_elem.get("name")
                filename = inputs_elem.get("filename")
                format = inputs_elem.get("format", "json")
                data_style = inputs_elem.get("data_style")
                content = {"format": format, "handle_files": data_style, "type": "inputs"}
                config_files.append(InputConfigFile(name=name, filename=filename, content=content))
        return config_files

    def parse_file_sources(self) -> Sequence[FileSourceConfigFile]:
        config_files: list[FileSourceConfigFile] = []
        if (conf_parent_elem := self.root.find("configfiles")) is not None:
            file_sources_elem = conf_parent_elem.find("file_sources")
            if file_sources_elem is not None:
                name = file_sources_elem.get("name")
                filename = file_sources_elem.get("filename", None)
                config_files.append(FileSourceConfigFile(name=name, filename=filename, content={"type": "files"}))
        return config_files

    def parse_creator(self):
        creators_el = self.root.find("creator")
        if creators_el is None:
            return None

        creators = []
        for creator_el in creators_el:
            creator_as_dict = {}
            if creator_el.tag == "person":
                clazz = "Person"
            elif creator_el.tag == "organization":
                clazz = "Organization"
            else:
                continue
            creator_as_dict["class"] = clazz
            creator_as_dict.update(creator_el.attrib)
            creators.append(creator_as_dict)
        return creators


def _test_elem_to_dict(test_elem, i, profile=None) -> ToolSourceTest:
    rval: ToolSourceTest = dict(
        outputs=__parse_output_elems(test_elem),
        output_collections=__parse_output_collection_elems(test_elem, profile=profile),
        inputs=__parse_input_elems(test_elem, i),
        expect_num_outputs=test_elem.get("expect_num_outputs"),
        command=__parse_assert_list_from_elem(test_elem.find("assert_command")),
        command_version=__parse_assert_list_from_elem(test_elem.find("assert_command_version")),
        stdout=__parse_assert_list_from_elem(test_elem.find("assert_stdout")),
        stderr=__parse_assert_list_from_elem(test_elem.find("assert_stderr")),
        expect_exit_code=test_elem.get("expect_exit_code"),
        expect_failure=string_as_bool(test_elem.get("expect_failure", False)),
        expect_test_failure=string_as_bool(test_elem.get("expect_test_failure", False)),
        maxseconds=test_elem.get("maxseconds", None),
    )
    _copy_to_dict_if_present(test_elem, rval, ["num_outputs"])
    return rval


def __parse_input_elems(test_elem, i) -> ToolSourceTestInputs:
    __expand_input_elems(test_elem)
    return __parse_inputs_elems(test_elem, i)


def __parse_output_elems(test_elem) -> ToolSourceTestOutputs:
    outputs: ToolSourceTestOutputs = []
    for output_elem in test_elem.findall("output"):
        name, file, attributes = __parse_output_elem(output_elem)
        outputs.append(ToolSourceTestOutput({"name": name, "value": file, "attributes": attributes}))
    return outputs


def __parse_output_elem(output_elem):
    attrib = _element_to_dict(output_elem)
    name = attrib.pop("name", None)
    if name is None:
        raise Exception("Test output does not have a 'name'")
    file, attributes = __parse_test_attributes(output_elem, attrib, parse_discovered_datasets=True)
    return name, file, attributes


def __parse_command_elem(test_elem):
    assert_elem = test_elem.find("command")
    return __parse_assert_list_from_elem(assert_elem)


def __parse_output_collection_elems(test_elem, profile=None):
    output_collections = []
    for output_collection_elem in test_elem.findall("output_collection"):
        output_collection_def = __parse_output_collection_elem(output_collection_elem, profile=profile)
        output_collections.append(output_collection_def)
    return output_collections


def __parse_output_collection_elem(output_collection_elem, profile=None):
    attrib = _element_to_dict(output_collection_elem)
    name = attrib.pop("name", None)
    if name is None:
        raise Exception("Test output collection does not have a 'name'")
    element_tests = __parse_element_tests(output_collection_elem, profile=profile)
    return TestCollectionOutputDef(name, attrib, element_tests).to_dict()


def __parse_element_tests(parent_element, profile=None):
    element_tests = {}
    for idx, element in enumerate(parent_element.findall("element")):
        element_attrib: dict = _element_to_dict(element)
        identifier = element_attrib.pop("name", None)
        if identifier is None:
            raise Exception("Test primary dataset does not have a 'identifier'")
        element_tests[identifier] = __parse_test_attributes(
            element, element_attrib, parse_elements=True, profile=profile
        )
        if profile and Version(profile) >= Version("20.09"):
            element_tests[identifier][1]["expected_sort_order"] = idx

    return element_tests


VALUE_OBJECT_UNSET = object()


def __parse_test_attributes(
    output_elem, attrib, parse_elements=False, parse_discovered_datasets=False, profile=None
) -> Tuple[Optional[str], ToolSourceTestOutputAttributes]:
    assert_list = __parse_assert_list(output_elem)

    # Allow either file or value to specify a target file to compare result with
    # file was traditionally used by outputs and value by extra files.
    file: Optional[str] = attrib.pop("file", attrib.pop("value", None))

    # File no longer required if an list of assertions was present.

    value_object: Any = VALUE_OBJECT_UNSET
    if "value_json" in attrib:
        value_object = json.loads(attrib.pop("value_json"))

    # Method of comparison
    compare: OutputCompareType = cast(OutputCompareType, attrib.pop("compare", "diff").lower())
    # Number of lines to allow to vary in logs (for dates, etc)
    lines_diff: int = int(attrib.pop("lines_diff", "0"))
    # Allow a file size to vary if sim_size compare
    delta: int = int(attrib.pop("delta", DEFAULT_DELTA))
    delta_frac: Optional[float] = float(attrib["delta_frac"]) if "delta_frac" in attrib else DEFAULT_DELTA_FRAC
    sort: bool = string_as_bool(attrib.pop("sort", DEFAULT_SORT))
    decompress: bool = string_as_bool(attrib.pop("decompress", DEFAULT_DECOMPRESS))
    # `location` may contain an URL to a remote file that will be used to download `file` (if not already present on disk).
    location: Optional[str] = attrib.get("location")
    if location and file is None:
        file = os.path.basename(location)  # If no file specified, try to get filename from URL last component
    # Parameters for "image_diff" comparison
    metric: str = attrib.pop("metric", DEFAULT_METRIC)
    eps: float = float(attrib.pop("eps", DEFAULT_EPS))
    pin_labels: Optional[Any] = attrib.pop("pin_labels", DEFAULT_PIN_LABELS)
    count: Optional[int] = None
    try:
        count = int(attrib.pop("count"))
    except KeyError:
        pass
    min: Optional[int] = None
    try:
        min = int(attrib.pop("min"))
    except KeyError:
        pass
    max: Optional[int] = None
    try:
        max = int(attrib.pop("max"))
    except KeyError:
        pass
    has_count_assertions = count is not None or min is not None or max is not None
    extra_files: List[Dict[str, Any]] = []
    ftype: Optional[str] = None
    if "ftype" in attrib:
        ftype = attrib["ftype"]
    for extra in output_elem.findall("extra_files"):
        extra_files.append(__parse_extra_files_elem(extra))
    metadata: Dict[str, Any] = {}
    for metadata_elem in output_elem.findall("metadata"):
        metadata[metadata_elem.get("name")] = metadata_elem.get("value")
    md5sum = attrib.get("md5", None)
    checksum = attrib.get("checksum", None)
    element_tests: Dict[str, Any] = {}
    if parse_elements:
        element_tests = __parse_element_tests(output_elem, profile=profile)

    primary_datasets: Dict[str, Any] = {}
    if parse_discovered_datasets:
        for primary_elem in output_elem.findall("discovered_dataset") or []:
            primary_attrib = _element_to_dict(primary_elem)
            designation = primary_attrib.pop("designation", None)
            if designation is None:
                raise Exception("Test primary dataset does not have a 'designation'")
            primary_datasets[designation] = __parse_test_attributes(primary_elem, primary_attrib)

    has_checksum = md5sum or checksum
    has_nested_tests = extra_files or element_tests or primary_datasets
    has_object = value_object is not VALUE_OBJECT_UNSET
    if not (
        assert_list
        or file
        or metadata
        or ftype
        or has_checksum
        or has_nested_tests
        or has_object
        or has_count_assertions
    ):
        raise Exception(
            "Test output defines nothing to check (e.g. must have a 'file' check against, assertions to check, metadata or checksum tests, etc...)"
        )
    attributes = ToolSourceTestOutputAttributes(
        ftype=ftype,
        compare=compare,
        lines_diff=lines_diff,
        delta=delta,
        delta_frac=delta_frac,
        sort=sort,
        decompress=decompress,
        metric=metric,
        eps=eps,
        pin_labels=pin_labels,
        location=location,
        count=count,
        min=min,
        max=max,
        metadata=metadata,
        md5=md5sum,
        checksum=checksum,
        primary_datasets=primary_datasets,
        elements=element_tests,
        assert_list=assert_list,
        extra_files=extra_files,
    )
    if value_object is not VALUE_OBJECT_UNSET:
        attributes["object"] = value_object
    return file, attributes


def __parse_assert_list(output_elem) -> AssertionList:
    assert_elem = output_elem.find("assert_contents")
    return __parse_assert_list_from_elem(assert_elem)


def __parse_assert_list_from_elem(assert_elem) -> AssertionList:
    assert_list = None

    def convert_elem(elem):
        """Converts and XML element to a dictionary format, used by assertion checking code."""
        tag = elem.tag
        attributes = _element_to_dict(elem)
        converted_children = []
        for child_elem in elem:
            converted_children.append(convert_elem(child_elem))
        return {"tag": tag, "attributes": attributes, "children": converted_children}

    if assert_elem is not None:
        assert_list = []
        for assert_child in list(assert_elem):
            assert_list.append(convert_elem(assert_child))

    return assert_list


def __parse_extra_files_elem(extra):
    # File or directory, when directory, compare basename
    # by basename
    attrib = _element_to_dict(extra)
    extra_type = attrib.pop("type", "file")
    extra_name = attrib.pop("name", None)
    assert (
        extra_type == "directory" or extra_name is not None
    ), f"extra_files type ({extra_type}) requires a name attribute"
    extra_value, extra_attributes = __parse_test_attributes(extra, attrib)
    return {"value": extra_value, "name": extra_name, "type": extra_type, "attributes": extra_attributes}


def __expand_input_elems(root_elem, prefix=""):
    __append_prefix_to_params(root_elem, prefix)

    repeat_elems = root_elem.findall("repeat")
    indices = {}
    for repeat_elem in repeat_elems:
        name = repeat_elem.get("name")
        if name not in indices:
            indices[name] = 0
            index = 0
        else:
            index = indices[name] + 1
            indices[name] = index

        new_prefix = __prefix_join(prefix, name, index=index)
        __expand_input_elems(repeat_elem, new_prefix)
        __pull_up_params(root_elem, repeat_elem)

    cond_elems = root_elem.findall("conditional")
    for cond_elem in cond_elems:
        new_prefix = __prefix_join(prefix, cond_elem.get("name"))
        __expand_input_elems(cond_elem, new_prefix)
        __pull_up_params(root_elem, cond_elem)

    section_elems = root_elem.findall("section")
    for section_elem in section_elems:
        new_prefix = __prefix_join(prefix, section_elem.get("name"))
        __expand_input_elems(section_elem, new_prefix)
        __pull_up_params(root_elem, section_elem)


def __append_prefix_to_params(elem, prefix):
    for param_elem in elem.findall("param"):
        param_elem.set("name", __prefix_join(prefix, param_elem.get("name")))


def __pull_up_params(parent_elem, child_elem):
    for param_elem in child_elem.findall("param"):
        parent_elem.append(param_elem)


def __prefix_join(prefix, name, index: Optional[int] = None):
    name = name if index is None else f"{name}_{index}"
    return name if not prefix else f"{prefix}|{name}"


def _copy_to_dict_if_present(elem, rval, attributes):
    for attribute in attributes:
        if attribute in elem.attrib:
            rval[attribute] = elem.get(attribute)
    return rval


def __parse_inputs_elems(test_elem, i) -> ToolSourceTestInputs:
    raw_inputs: ToolSourceTestInputs = []
    for param_elem in test_elem.findall("param"):
        raw_inputs.append(__parse_param_elem(param_elem, i))

    return raw_inputs


def _test_collection_def_dict(elem: Element) -> XmlTestCollectionDefDict:
    elements: List[TestCollectionDefElementDict] = []
    attrib: Dict[str, Any] = _element_to_dict(elem)
    collection_type = attrib["type"]
    name = attrib.get("name", "Unnamed Collection")
    for element in elem.findall("element"):
        element_attrib: Dict[str, Any] = _element_to_dict(element)
        element_identifier = element_attrib["name"]
        nested_collection_elem = element.find("collection")
        element_definition: TestCollectionDefElementObject
        if nested_collection_elem is not None:
            element_definition = _test_collection_def_dict(nested_collection_elem)
        else:
            element_definition = __parse_param_elem(element)
        elements.append({"element_identifier": element_identifier, "element_definition": element_definition})
    fields_json = "null"
    fields_el = elem.find("fields")
    if fields_el is not None:
        fields_json = fields_el.text or "null"
    fields = json.loads(fields_json)
    return XmlTestCollectionDefDict(
        model_class="TestCollectionDef",
        attributes=attrib,
        collection_type=collection_type,
        elements=elements,
        name=name,
        fields=fields,
    )


def __parse_param_elem(param_elem, i=0) -> ToolSourceTestInput:
    attrib: ToolSourceTestInputAttributes = dict(param_elem.attrib)
    if "values" in attrib:
        value = attrib["values"].split(",")
    elif "value" in attrib:
        value = attrib["value"]
    elif "value_json" in attrib:
        value = json.loads(attrib["value_json"])
    else:
        value = None

    if value is None and attrib.get("location", None) is not None:
        value = os.path.basename(attrib["location"])

    children_elem = param_elem
    if children_elem is not None:
        # At this time, we can assume having children only
        # occurs on DataToolParameter test items but this could
        # change and would cause the below parsing to change
        # based upon differences in children items
        attrib["metadata"] = {}
        attrib["composite_data"] = []
        attrib["edit_attributes"] = []
        # Composite datasets need to be renamed uniquely
        composite_data_name = None
        for child in children_elem:
            if child.tag == "composite_data":
                file_name = child.get("value")
                attrib["composite_data"].append(file_name)
                if composite_data_name is None:
                    # Generate a unique name; each test uses a
                    # fresh history.
                    composite_data_name = f"_COMPOSITE_RENAMED_t{i}_{uuid.uuid1().hex}"
            elif child.tag == "metadata":
                attrib["metadata"][child.get("name")] = child.get("value")
            elif child.tag == "edit_attributes":
                attrib["edit_attributes"].append(child)
            elif child.tag == "collection":
                attrib["collection"] = _test_collection_def_dict(child)
        if composite_data_name:
            # Composite datasets need implicit renaming;
            # inserted at front of list so explicit declarations
            # take precedence
            attrib["edit_attributes"].insert(0, {"type": "name", "value": composite_data_name})
    name = attrib.pop("name")
    return {"name": name, "value": value, "attributes": attrib}


class StdioParser:
    def __init__(self, root):
        try:
            self.stdio_exit_codes = []
            self.stdio_regexes = []

            # We should have a single <stdio> element, but handle the case for
            # multiples.
            # For every stdio element, add all of the exit_code and regex
            # subelements that we find:
            for stdio_elem in root.findall("stdio"):
                self.parse_stdio_exit_codes(stdio_elem)
                self.parse_stdio_regexes(stdio_elem)
        except Exception:
            log.exception("Exception in parse_stdio!")

    def parse_stdio_exit_codes(self, stdio_elem):
        """
        Parse the tool's <stdio> element's <exit_code> subelements.
        This will add all of those elements, if any, to self.stdio_exit_codes.
        """
        try:
            # Look for all <exit_code> elements. Each exit_code element must
            # have a range/value.
            # Exit-code ranges have precedence over a single exit code.
            # So if there are value and range attributes, we use the range
            # attribute. If there is neither a range nor a value, then print
            # a warning and skip to the next.
            for exit_code_elem in stdio_elem.findall("exit_code"):
                exit_code = ToolStdioExitCode()
                # Each exit code has an optional description that can be
                # part of the "desc" or "description" attributes:
                exit_code.desc = exit_code_elem.get("desc")
                if exit_code.desc is None:
                    exit_code.desc = exit_code_elem.get("description")
                # Parse the error level:
                exit_code.error_level = self.parse_error_level(exit_code_elem.get("level"))
                code_range = exit_code_elem.get("range")
                if code_range is None:
                    code_range = exit_code_elem.get("value")
                if code_range is None:
                    log.warning("Tool stdio exit codes must have a range or value")
                    continue
                # Parse the range. We look for:
                #   :Y
                #  X:
                #  X:Y   - Split on the colon. We do not allow a colon
                #          without a beginning or end, though we could.
                # Also note that whitespace is eliminated.
                # TODO: Turn this into a single match - it should be
                # more efficient.
                code_range = re.sub(r"\s", "", code_range)
                code_ranges = re.split(r":", code_range)
                if len(code_ranges) == 2:
                    if code_ranges[0] is None or "" == code_ranges[0]:
                        exit_code.range_start = -math.inf
                    else:
                        exit_code.range_start = int(code_ranges[0])
                    if code_ranges[1] is None or "" == code_ranges[1]:
                        exit_code.range_end = math.inf
                    else:
                        exit_code.range_end = int(code_ranges[1])
                # If we got more than one colon, then ignore the exit code.
                elif len(code_ranges) > 2:
                    log.warning(f"Invalid tool exit_code range {code_range} - ignored")
                    continue
                # Else we have a singular value. If it's not an integer, then
                # we'll just write a log message and skip this exit_code.
                else:
                    try:
                        exit_code.range_start = int(code_range)
                    except Exception:
                        log.error(code_range)
                        log.warning(f"Invalid range start for tool's exit_code {code_range}: exit_code ignored")
                        continue
                    exit_code.range_end = exit_code.range_start
                # TODO: Check if we got ">", ">=", "<", or "<=":
                # Check that the range, regardless of how we got it,
                # isn't bogus. If we have two infinite values, then
                # the start must be -inf and the end must be +inf.
                # So at least warn about this situation:
                if math.isinf(exit_code.range_start) and math.isinf(exit_code.range_end):
                    log.warning(f"Tool exit_code range {code_range} will match on all exit codes")
                self.stdio_exit_codes.append(exit_code)
        except Exception:
            log.exception("Exception in parse_stdio_exit_codes!")

    def parse_stdio_regexes(self, stdio_elem):
        """
        Look in the tool's <stdio> elem for all <regex> subelements
        that define how to look for warnings and fatal errors in
        stdout and stderr. This will add all such regex elements
        to the Tols's stdio_regexes list.
        """
        try:
            # Look for every <regex> subelement. The regular expression
            # will have "match" and "source" (or "src") attributes.
            for regex_elem in stdio_elem.findall("regex"):
                # TODO: Fill in ToolStdioRegex
                regex = ToolStdioRegex()
                # Each regex has an optional description that can be
                # part of the "desc" or "description" attributes:
                regex.desc = regex_elem.get("desc")
                if regex.desc is None:
                    regex.desc = regex_elem.get("description")
                # Parse the error level
                regex.error_level = self.parse_error_level(regex_elem.get("level"))
                regex.match = regex_elem.get("match")
                if regex.match is None:
                    log.warning(
                        f"Ignoring tool's stdio regex element with attributes {regex_elem.attrib} - "
                        "the 'match' attribute must exist"
                    )
                    continue
                # Parse the output sources. We look for the "src", "source",
                # and "sources" attributes, in that order. If there is no
                # such source, then the source defaults to stderr & stdout.
                # Look for a comma and then look for "err", "error", "out",
                # and "output":
                output_srcs = regex_elem.get("src")
                if output_srcs is None:
                    output_srcs = regex_elem.get("source")
                if output_srcs is None:
                    output_srcs = regex_elem.get("sources")
                if output_srcs is None:
                    output_srcs = "output,error"
                output_srcs = re.sub(r"\s", "", output_srcs)
                src_list = re.split(r",", output_srcs)
                # Just put together anything to do with "out", including
                # "stdout", "output", etc. Repeat for "stderr", "error",
                # and anything to do with "err". If neither stdout nor
                # stderr were specified, then raise a warning and scan both.
                for src in src_list:
                    if re.search("both", src, re.IGNORECASE):
                        regex.stdout_match = True
                        regex.stderr_match = True
                    if re.search("out", src, re.IGNORECASE):
                        regex.stdout_match = True
                    if re.search("err", src, re.IGNORECASE):
                        regex.stderr_match = True
                    if not regex.stdout_match and not regex.stderr_match:
                        log.warning(
                            "Tool id %s: unable to determine if tool "
                            "stream source scanning is output, error, "
                            "or both. Defaulting to use both.",
                            self.id,
                        )
                        regex.stdout_match = True
                        regex.stderr_match = True
                self.stdio_regexes.append(regex)
        except Exception:
            log.exception("Exception in parse_stdio_exit_codes!")

    # TODO: This method doesn't have to be part of the Tool class.
    def parse_error_level(self, err_level):
        """
        Parses error level and returns error level enumeration. If
        unparsable, returns 'fatal'
        """
        return_level = StdioErrorLevel.FATAL
        try:
            if err_level:
                if re.search("log", err_level, re.IGNORECASE):
                    return_level = StdioErrorLevel.LOG
                elif re.search("qc", err_level, re.IGNORECASE):
                    return_level = StdioErrorLevel.QC
                elif re.search("warning", err_level, re.IGNORECASE):
                    return_level = StdioErrorLevel.WARNING
                elif re.search("fatal_oom", err_level, re.IGNORECASE):
                    return_level = StdioErrorLevel.FATAL_OOM
                elif re.search("fatal", err_level, re.IGNORECASE):
                    return_level = StdioErrorLevel.FATAL
                else:
                    log.debug(f"Tool {self.id}: error level {err_level} did not match log/warning/fatal")
        except Exception:
            log.exception("Exception in parse_error_level")
        return return_level


class XmlPagesSource(PagesSource):
    def __init__(self, root):
        self.input_elem = root.find("inputs")
        page_sources = []
        if self.input_elem is not None:
            pages_elem = self.input_elem.findall("page")
            for page in pages_elem or [self.input_elem]:
                page_sources.append(XmlPageSource(page))
        super().__init__(page_sources)

    @property
    def inputs_defined(self):
        return self.input_elem is not None


class XmlPageSource(PageSource):
    def __init__(self, parent_elem):
        self.parent_elem = parent_elem

    def parse_display(self):
        display_elem = self.parent_elem.find("display")
        if display_elem is not None:
            display = xml_to_string(display_elem)
        else:
            display = None
        return display

    def parse_input_sources(self):
        return list(map(XmlInputSource, self.parent_elem))


class XmlDynamicOptions(DynamicOptions):

    def __init__(self, options_elem: Element, dynamic_option_code: Optional[str]):
        self._options_elem = options_elem
        self._dynamic_options_code = dynamic_option_code

    def elem(self) -> Element:
        return self._options_elem

    def get_dynamic_options_code(self) -> Optional[str]:
        """If dynamic options are a piece of code to eval, return it."""
        return self._dynamic_options_code

    def get_data_table_name(self) -> Optional[str]:
        """If dynamic options are loaded from a data table, return the name."""
        return self._options_elem.get("from_data_table") if self._options_elem is not None else None

    def get_index_file_name(self) -> Optional[str]:
        return self._options_elem.get("from_file") if self._options_elem is not None else None


class XmlInputSource(InputSource):
    def __init__(self, input_elem):
        self.input_elem = input_elem
        self.input_type = self.input_elem.tag

    def parse_input_type(self):
        return self.input_type

    def parse_extensions(self):
        return [extension.strip().lower() for extension in self.get("format", "data").split(",")]

    def elem(self):
        return self.input_elem

    def get(self, key, value=None):
        return self.input_elem.get(key, value)

    def get_bool(self, key, default):
        return string_as_bool(self.get(key, default))

    def get_bool_or_none(self, key, default):
        return string_as_bool_or_none(self.get(key, default))

    def parse_label(self):
        return xml_text(self.input_elem, "label")

    def parse_help(self):
        return xml_text(self.input_elem, "help")

    def parse_sanitizer_elem(self):
        return self.input_elem.find("sanitizer")

    def parse_validators(self) -> List[AnyValidatorModel]:
        return parse_xml_validators(self.input_elem)

    def parse_dynamic_options(self) -> Optional[XmlDynamicOptions]:
        """Return a XmlDynamicOptions to describe dynamic options if options elem is available."""
        options_elem = self.input_elem.find("options")
        dynamic_option_code = self.input_elem.get("dynamic_options")
        is_dynamic = options_elem is not None or dynamic_option_code is not None
        if is_dynamic:
            return XmlDynamicOptions(options_elem, dynamic_option_code)
        else:
            return None

    def parse_static_options(self) -> List[Tuple[str, str, bool]]:
        """
        >>> from galaxy.util import parse_xml_string_to_etree
        >>> xml = '<param><option value="a">A</option><option value="b">B</option></param>'
        >>> xis = XmlInputSource(parse_xml_string_to_etree(xml).getroot())
        >>> xis.parse_static_options()
        [('A', 'a', False), ('B', 'b', False)]
        >>> xml = '<param><option value="a"/><option value="b"/><option value="a" selected="true"/></param>'
        >>> xis = XmlInputSource(parse_xml_string_to_etree(xml).getroot())
        >>> xis.parse_static_options()
        [('a', 'a', True), ('b', 'b', False)]
        """

        deduplicated_static_options = {}

        elem = self.input_elem
        for option in elem.findall("option"):
            value = option.get("value")
            text = option.text or value
            selected = string_as_bool(option.get("selected", False))
            deduplicated_static_options[value] = (text, value, selected)
        return list(deduplicated_static_options.values())

    def parse_drill_down_dynamic_options(
        self, tool_data_path: Optional[str] = None
    ) -> Optional[DrillDownDynamicOptions]:
        elem = self.input_elem
        dynamic_options_raw = elem.get("dynamic_options", None)
        dynamic_options: Optional[str] = str(dynamic_options_raw) if dynamic_options_raw else None
        if dynamic_options is None:
            return None
        else:
            return XmlDrillDownDynamicOptions(code_block=dynamic_options)

    def parse_drill_down_static_options(
        self, tool_data_path: Optional[str] = None
    ) -> Optional[List[DrillDownOptionsDict]]:
        from_file = self.input_elem.get("from_file", None)
        if from_file:
            if not os.path.isabs(from_file):
                assert tool_data_path, "This tool cannot be parsed outside of a Galaxy context"
                from_file = os.path.join(tool_data_path, from_file)
            elem = XML(f"<root>{open(from_file).read()}</root>")
        else:
            elem = self.input_elem

        dynamic_options_elem = elem.get("dynamic_options", None)
        filter_elem = elem.get("filter", None)
        if dynamic_options_elem is not None and filter_elem is not None:
            return None

        root_options: List[DrillDownOptionsDict] = []
        options_elem = elem.find("options")
        assert options_elem, "Non-dynamic drilldown parameters must supply an options element"
        _recurse_drill_down_elems(root_options, options_elem.findall("option"))
        return root_options

    def parse_optional(self, default=None):
        """Return boolean indicating whether parameter is optional."""
        elem = self.input_elem
        if self.get("type") == "data_column":
            # Allow specifing force_select for backward compat., but probably
            # should use optional going forward for consistency with other
            # parameters.
            if "force_select" in elem.attrib:
                force_select = string_as_bool(elem.get("force_select"))
            else:
                force_select = not string_as_bool(elem.get("optional", False))
            return not force_select

        if default is None:
            default = self.default_optional
        return self.get_bool("optional", default)

    def parse_conversion_tuples(self):
        elem = self.input_elem
        conversions = []
        for conv_elem in elem.findall("conversion"):
            name = conv_elem.get("name")  # name for commandline substitution
            conv_extensions = conv_elem.get("type")  # target datatype extension
            conversions.append((name, conv_extensions))
        return conversions

    def parse_nested_inputs_source(self):
        elem = self.input_elem
        return XmlPageSource(elem)

    def parse_test_input_source(self):
        elem = self.input_elem
        input_elem = elem.find("param")
        assert input_elem is not None, "<conditional> must have a child <param>"
        return XmlInputSource(input_elem)

    def parse_when_input_sources(self):
        elem = self.input_elem

        sources = []
        for case_elem in elem.findall("when"):
            value = case_elem.get("value")
            case_page_source = XmlPageSource(case_elem)
            sources.append((value, case_page_source))
        return sources

    def parse_default(self) -> Optional[Dict[str, Any]]:
        def file_default_from_elem(elem):
            # TODO: hashes, created_from_basename, etc...
            return {"class": "File", "location": elem.get("location")}

        def read_elements(collection_elem):
            element_dicts = []
            elements = collection_elem.findall("element")
            for element in elements:
                identifier = element.get("name")
                subcollection_elem = element.find("collection")
                if subcollection_elem:
                    collection_type = subcollection_elem.get("collection_type")
                    element_dicts.append(
                        {
                            "class": "Collection",
                            "identifier": identifier,
                            "collection_type": collection_type,
                            "elements": read_elements(subcollection_elem),
                        }
                    )
                else:
                    element_dict = file_default_from_elem(element)
                    element_dict["identifier"] = identifier
                    element_dicts.append(element_dict)
            return element_dicts

        elem = self.input_elem
        element_type = self.input_elem.get("type")
        if element_type == "data":
            default_elem = elem.find("default")
            if default_elem is not None:
                return file_default_from_elem(default_elem)
            else:
                return None
        else:
            default_elem = elem.find("default")
            if default_elem is not None:
                default_elem = elem.find("default")
                collection_type = default_elem.get("collection_type")
                name = default_elem.get("name", elem.get("name"))
                return {
                    "class": "Collection",
                    "name": name,
                    "collection_type": collection_type,
                    "elements": read_elements(default_elem),
                }
            else:
                return None


class ParallelismInfo:
    """
    Stores the information (if any) for running multiple instances of the tool in parallel
    on the same set of inputs.
    """

    def __init__(self, tag):
        self.method = tag.get("method")
        if isinstance(tag, dict):
            items = tag.items()
        else:
            items = tag.attrib.items()
        self.attributes = dict([item for item in items if item[0] != "method"])
        if len(self.attributes) == 0:
            # legacy basic mode - provide compatible defaults
            self.attributes["split_size"] = 20
            self.attributes["split_mode"] = "number_of_parts"


def parse_citation_elem(citation_elem: Element) -> Optional[Citation]:
    if citation_elem.tag != "citation":
        return None

    citation_type = citation_elem.attrib.get("type", None)
    citation_raw_text = citation_elem.text
    assert citation_raw_text
    content = citation_raw_text.strip()

    return Citation(
        type=citation_type,
        content=content,
    )


class XmlDrillDownDynamicOptions(DrillDownDynamicOptions):

    def __init__(self, code_block: Optional[str]):
        self._code_block = code_block

    def from_code_block(self) -> Optional[str]:
        """Get a code block to do an eval on."""
        return self._code_block


def _element_to_dict(elem: Element) -> Dict[str, Any]:
    # every call to this function needs to be replaced with something more type safe and with
    # an actual typed dictionary - but centralizing this hack for now.
    return dict(elem.attrib)  # type: ignore [arg-type]


def _recurse_drill_down_elems(options: List[DrillDownOptionsDict], option_elems: List[Element]):
    for option_elem in option_elems:
        selected = string_as_bool(option_elem.get("selected", False))
        nested_options: List[DrillDownOptionsDict] = []
        value = option_elem.get("value")
        assert value
        current_option: DrillDownOptionsDict = DrillDownOptionsDict(
            {
                "name": option_elem.get("name"),
                "value": value,
                "options": nested_options,
                "selected": selected,
            }
        )
        _recurse_drill_down_elems(nested_options, option_elem.findall("option"))
        options.append(current_option)



#!/usr/bin/env python

# how to use this function...
# PYTHONPATH=lib python lib/galaxy/tool_util/verify/codegen.py

import argparse
import inspect
import os
from shutil import move
from typing import (
    cast,
    List,
    Optional,
    Union,
)

import lxml.etree as ET
from jinja2 import Environment
from typing_extensions import (
    Annotated,
    get_args,
    get_origin,
    Literal,
)

from galaxy.tool_util.verify.asserts import assertion_module_and_functions
from galaxy.tool_util.verify.asserts._types import AssertionParameter as AssertionParameterAnnotation
from galaxy.util.commands import shell

models_path = os.path.join(os.path.dirname(__file__), "assertion_models.py")
galaxy_xsd_path = os.path.join(os.path.dirname(__file__), "..", "xsd", "galaxy.xsd")

Children = Literal["allowed", "required", "forbidden"]

DESCRIPTION = """This script synchronizes dynamic code artifacts against models in Galaxy.

Right now this just synchronizes Galaxy's XSD file against documentation in Galaxy's assertion modules but
in the future it will also build Pydantic models for these functions.
"""

assert_models_template = """
'''This module is auto-generated, please do not modify.'''
import typing
import re

from typing_extensions import (
    Annotated,
    Literal,
)

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    model_validator,
    RootModel,
    StrictFloat,
    StrictInt,
)

BYTES_PATTERN = re.compile(r"^(0|[1-9][0-9]*)([kKMGTPE]i?)?$")


class AssertionModel(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )


def check_bytes(v: typing.Any) -> typing.Any:
    if isinstance(v, str):
        assert BYTES_PATTERN.match(v), "Not valid bytes string"
    return v


def check_center_of_mass(v: typing.Any):
    assert isinstance(v, str)
    split_parts = v.split(",")
    assert len(split_parts) == 2
    for part in split_parts:
        assert float(part.strip())
    return v


def check_regex(v: typing.Any):
    assert isinstance(v, str)
    try:
        re.compile(v)
    except re.error:
        raise AssertionError(f"Invalid regular expression {v}")
    return v


def check_non_negative_if_set(v: typing.Any):
    if v is not None:
        try:
            assert float(v) >= 0
        except TypeError:
            raise AssertionError(f"Invalid type found {v}")
    return v


def check_non_negative_if_int(v: typing.Any):
    if v is not None and isinstance(v, int):
        assert v >= 0
    return v


{% for assertion in assertions %}
{% for parameter in assertion.parameters %}
{{assertion.name}}_{{ parameter.name }}_description = '''{{ parameter.description }}'''
{% endfor %}

class base_{{assertion.name}}_model(AssertionModel):
    '''base model for {{assertion.name}} describing attributes.'''
{% for parameter in assertion.parameters %}
{% if not parameter.is_deprecated %}
    {{ parameter.name }}: {{ parameter.type_str }} = Field(
        {{ parameter.field_default_str }},
        description={{ assertion.name }}_{{ parameter.name }}_description,
    )
{% endif %}
{% endfor %}
{% if assertion.children in ["required", "allowed"] %}
    children: typing.Optional["assertion_list"] = None
    asserts: typing.Optional["assertion_list"] = None

{% if assertion.children == "required" %}
    @model_validator(mode='before')
    @classmethod
    def validate_children(self, data: typing.Any):
        if isinstance(data, dict) and 'children' not in data and 'asserts' not in data:
            raise ValueError("At least one of 'children' or 'asserts' must be specified for this assertion type.")
        return data
{% endif %}
{% endif %}


class base_{{assertion.name}}_model_relaxed(AssertionModel):
    '''base model for {{assertion.name}} describing attributes.'''
{% for parameter in assertion.parameters %}
{% if not parameter.is_deprecated %}
    {{ parameter.name }}: {{ parameter.lax_type_str }} = Field(
        {{ parameter.field_default_str }},
        description={{ assertion.name }}_{{ parameter.name }}_description,
    )
{% endif %}
{% endfor %}
{% if assertion.children in ["required", "allowed"] %}
    children: typing.Optional["assertion_list"] = None
    asserts: typing.Optional["assertion_list"] = None

{% if assertion.children == "required" %}
    @model_validator(mode='before')
    @classmethod
    def validate_children(self, data: typing.Any):
        if isinstance(data, dict) and 'children' not in data and 'asserts' not in data:
            raise ValueError("At least one of 'children' or 'asserts' must be specified for this assertion type.")
        return data
{% endif %}
{% endif %}


class {{assertion.name}}_model(base_{{assertion.name}}_model):
    r\"\"\"{{ assertion.docstring }}\"\"\"
    that: Literal["{{assertion.name}}"] = "{{assertion.name}}"

class {{assertion.name}}_model_nested(AssertionModel):
    r\"\"\"Nested version of this assertion model.\"\"\"
    {{assertion.name}}: base_{{assertion.name}}_model

class {{assertion.name}}_model_relaxed(base_{{assertion.name}}_model_relaxed):
    r\"\"\"{{ assertion.docstring }}\"\"\"
    that: Literal["{{assertion.name}}"] = "{{assertion.name}}"
{% endfor %}

any_assertion_model_flat = Annotated[typing.Union[
{% for assertion in assertions %}
    {{assertion.name}}_model,
{% endfor %}
], Field(discriminator="that")]

any_assertion_model_nested = typing.Union[
{% for assertion in assertions %}
    {{assertion.name}}_model_nested,
{% endfor %}
]

any_assertion_model_flat_relaxed = Annotated[typing.Union[
{% for assertion in assertions %}
    {{assertion.name}}_model_relaxed,
{% endfor %}
], Field(discriminator="that")]

assertion_list = RootModel[typing.List[typing.Union[any_assertion_model_flat, any_assertion_model_nested]]]

# used to model what the XML conversion should look like - not meant to be consumed outside of
# of Galaxy internals / linting.
relaxed_assertion_list = RootModel[typing.List[any_assertion_model_flat_relaxed]]

class assertion_dict(AssertionModel):
{% for assertion in assertions %}
    {{assertion.name}}: typing.Optional[base_{{assertion.name}}_model] = None
{% endfor %}


assertions = typing.Union[assertion_list, assertion_dict]
"""


def get_default_args(func):
    signature = inspect.signature(func)
    return {k: v.default for k, v in signature.parameters.items()}


def main():
    assertions = []
    for function_name, (module_and_function, assertion_function) in assertion_module_and_functions.items():
        if not function_name.startswith("assert_"):
            continue
        name = function_name[len("assert_") :]
        docstring = inspect.cleandoc(assertion_function.__doc__ or "")
        annotations = assertion_function.__annotations__
        default_args = get_default_args(assertion_function)
        parameters = []
        children: Children = "forbidden"
        for parameter_name, parameter_type in annotations.items():
            if parameter_name == "return":
                continue
            elif parameter_name == "children":
                if default_args.get("children", inspect._empty) is not inspect._empty:
                    children = "allowed"
                else:
                    children = "required"
                continue
            elif parameter_name in ["output", "output_bytes", "verify_assertions_function"]:
                continue

            default_value = default_args.get(parameter_name)
            parameters.append(AssertionParameter(parameter_name, parameter_type, default_value))
        assertion = Assertion(name, docstring, parameters, children, module_and_function)
        assertions.append(assertion)
    rewrite_galaxy_xsd(assertions)
    write_assertion_models(assertions)


def _expand_template(template_str: str, assertions) -> str:
    template = Environment().from_string(template_str)
    return template.render(assertions=assertions)


def write_assertion_models(assertions):
    models_file_contents = _expand_template(assert_models_template, assertions)
    with open(models_path, "w") as f:
        f.write(models_file_contents)
    shell(["isort", models_path])
    shell(["black", models_path])


def get_annotation(text: str, nsmap):
    annotation = ET.Element("{http://www.w3.org/2001/XMLSchema}annotation")
    documentation = ET.Element("{http://www.w3.org/2001/XMLSchema}documentation")
    documentation.text = ET.CDATA(text)
    documentation.attrib["{http://www.w3.org/XML/1998/namespace}lang"] = "en"
    annotation.append(documentation)
    return annotation


def parameter_xsd_element(assertion_parameter, nsmap):
    el = ET.Element("{http://www.w3.org/2001/XMLSchema}attribute", nsmap=nsmap)
    el.attrib["name"] = assertion_parameter.name
    el.attrib["type"] = assertion_parameter.xml_type_str
    el.attrib["use"] = "required" if not assertion_parameter.has_default_value else "optional"
    annotation = get_annotation(assertion_parameter.description, nsmap)
    el.append(annotation)
    return el


def xsd_element(assertion, nsmap):
    el = ET.Element("{http://www.w3.org/2001/XMLSchema}element", nsmap=nsmap)
    el.attrib["name"] = assertion.name
    annotation = get_annotation(assertion.docstring + "\n\n$attribute_list::5", nsmap)
    complexType = ET.Element("{http://www.w3.org/2001/XMLSchema}complexType", nsmap=nsmap)
    if assertion.children != "forbidden":
        sequence = ET.Element("{http://www.w3.org/2001/XMLSchema}sequence", nsmap=nsmap)
        group = ET.Element("{http://www.w3.org/2001/XMLSchema}group", nsmap=nsmap)
        group.attrib["ref"] = "TestAssertion"
        group.attrib["minOccurs"] = "0" if assertion.children == "allowed" else "1"
        group.attrib["maxOccurs"] = "unbounded"
        sequence.append(group)
        complexType.append(sequence)
    for parameter in assertion.parameters:
        complexType.append(parameter_xsd_element(parameter, nsmap))
    el.append(annotation)
    el.append(complexType)
    return el


def xsd_elements(assertions, nsmap):
    elements = []
    for assertion in assertions:
        elements.append(ET.Comment(f" XSD doc for element auto-generated from {assertion.module_and_function}"))
        elements.append(xsd_element(assertion, nsmap))
    return elements


def rewrite_galaxy_xsd(assertions):
    with open(galaxy_xsd_path, "rb") as f:
        contents = f.read()
    parser = ET.XMLParser(strip_cdata=False)
    root = ET.fromstring(contents, parser=parser)
    raw_element = root.find(
        ".//{http://www.w3.org/2001/XMLSchema}group[@name='TestAssertion']/{http://www.w3.org/2001/XMLSchema}choice"
    )
    assert raw_element
    element = cast(ET._Element, raw_element)
    for el in element.iterchildren():
        element.remove(el)
    element.append(ET.Comment("The following block and all children are auto-generated - please do not modify."))
    for xsd_element in xsd_elements(assertions, root.nsmap):
        xsd_element.tail = "\n      "
        element.append(xsd_element)
    xml_new = ET.tostring(root).decode("utf-8")
    with open(galaxy_xsd_path, "w") as f:
        f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        f.write(xml_new)
        f.write("\n")
    ret_code = shell(["xmllint", "--format", "--output", "galaxy-tmp.xsd", galaxy_xsd_path])
    assert ret_code == 0
    move("galaxy-tmp.xsd", galaxy_xsd_path)


class AssertionParameter:

    def __init__(self, name: str, type: str, default_value):
        self.name = name
        self.type = type
        self.default_value = default_value

    @property
    def description(self) -> str:
        type = self.type
        if hasattr(type, "__metadata__"):
            return type.__metadata__[0].doc
        else:
            return ""

    @property
    def type_str(self) -> str:
        raw_type_str = as_type_str(self.type, strict=True)
        validators = self.validators[:]
        if self.xml_type_str == "Bytes":
            validators.append("check_bytes")
            validators.append("check_non_negative_if_int")
        if len(validators) > 0:
            validation_str = ",".join([f"BeforeValidator({v})" for v in validators])
            return f"Annotated[{raw_type_str}, {validation_str}]"

        return raw_type_str

    @property
    def lax_type_str(self) -> str:
        raw_type_str = as_type_str(self.type, strict=False)
        validators = self.validators[:]
        if self.xml_type_str == "Bytes":
            validators.append("check_bytes")
            validators.append("check_non_negative_if_int")
        if len(validators) > 0:
            validation_str = ",".join([f"BeforeValidator({v})" for v in validators])
            return f"Annotated[{raw_type_str}, {validation_str}]"

        return raw_type_str

    @property
    def xml_type_str(self) -> str:
        return as_xml_type(self.type)

    @property
    def field_default_str(self) -> str:
        if not self.has_default_value:
            return "..."
        elif isinstance(self.default_value, str):
            return f'''"{self.default_value}"'''
        else:
            return str(self.default_value)

    @property
    def has_default_value(self) -> bool:
        return self.default_value is not inspect._empty

    @property
    def is_deprecated(self) -> bool:
        assertion_parameter = self._get_type_annotation()
        if assertion_parameter is not None:
            return assertion_parameter.deprecated

        return False

    @property
    def validators(self) -> List[str]:
        assertion_parameter = self._get_type_annotation()
        if assertion_parameter is not None:
            return assertion_parameter.validators

        return []

    def _get_type_annotation(self) -> Optional[AssertionParameterAnnotation]:
        target_type = self.type
        if get_origin(target_type) is Annotated:
            args = get_args(target_type)
            if len(args) > 1:
                return cast(AssertionParameterAnnotation, args[1])

        return None


def _is_none_type(target_type):
    return target_type is type(None)


def _non_optional_types(union_type):
    return [t for t in get_args(union_type) if not _is_none_type(t)]


def as_xml_type(target_type) -> str:
    if get_origin(target_type) is Annotated:
        args = get_args(target_type)
        if len(args) > 1:
            assertion_parameter = args[1]
            if assertion_parameter.xml_type:
                return assertion_parameter.xml_type

        return as_xml_type(args[0])
    elif get_origin(target_type) is Union:
        types = _non_optional_types(target_type)
        if len(types) == 2:
            non_str_types = [t for t in types if t is not str]
            if len(non_str_types) == 1 and non_str_types[0] is bool:
                return "xs:boolean"
            elif len(non_str_types) == 1 and non_str_types[0] is int:
                return "xs:integer"
            elif len(non_str_types) == 1 and non_str_types[0] is float:
                return "xs:float"

    return "xs:string"


def as_type_str(target_type, strict=True):
    if get_origin(target_type) is Annotated:
        args = get_args(target_type)
        if len(args) > 1:
            if args[1].json_type and strict:
                return args[1].json_type

        return as_type_str(args[0])
    elif get_origin(target_type) is Union:
        is_optional = any(_is_none_type(t) for t in get_args(target_type))
        types_as_str = ", ".join(map(as_type_str, _non_optional_types(target_type)))
        union_type = f"typing.Union[{types_as_str}]"
        if is_optional:
            return f"typing.Optional[{union_type}]"
        else:
            return union_type
    elif target_type is str:
        return "str"
    elif target_type is int:
        return "int"
    elif target_type is float:
        return "float"
    elif target_type is bool:
        return "bool"
    else:
        return str(target_type)


class Assertion:

    def __init__(
        self,
        name: str,
        docstring: str,
        parameters: List[AssertionParameter],
        children: Children,
        module_and_function: str,
    ):
        self.name = name
        self.parameters = parameters
        self.docstring = docstring
        self.children = children
        self.module_and_function = module_and_function


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    return parser


if __name__ == "__main__":
    main()



"""This module contains linting functions for general aspects of the tool."""

import re
from typing import (
    Tuple,
    TYPE_CHECKING,
)

from packaging.version import Version

from galaxy.tool_util.biotools.source import ApiBiotoolsMetadataSource
from galaxy.tool_util.edam_util import load_edam_tree
from galaxy.tool_util.lint import Linter
from galaxy.tool_util.version import (
    LegacyVersion,
    parse_version,
)

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser.interface import ToolSource
    from galaxy.util.etree import (
        Element,
        ElementTree,
    )

PROFILE_PATTERN = re.compile(r"^[12]\d\.\d{1,2}$")


lint_tool_types = ["*"]


def _tool_xml_and_root(tool_source: "ToolSource") -> Tuple["ElementTree", "Element"]:
    tool_xml = getattr(tool_source, "xml_tree", None)
    if tool_xml:
        tool_node = tool_xml.getroot()
    else:
        tool_node = None
    return tool_xml, tool_node


class ToolVersionMissing(Linter):
    """
    Tools must have a version
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml, tool_node = _tool_xml_and_root(tool_source)
        version = tool_source.parse_version() or ""
        if not version:
            lint_ctx.error("Tool version is missing or empty.", linter=cls.name(), node=tool_node)


class ToolVersionPEP404(Linter):
    """
    Tools should have a PEP404 compliant version.
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml, tool_node = _tool_xml_and_root(tool_source)
        version = tool_source.parse_version() or ""
        parsed_version = parse_version(version)
        if version and isinstance(parsed_version, LegacyVersion):
            lint_ctx.warn(f"Tool version [{version}] is not compliant with PEP 440.", linter=cls.name(), node=tool_node)


class ToolVersionWhitespace(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml, tool_node = _tool_xml_and_root(tool_source)
        version = tool_source.parse_version() or ""
        if version != version.strip():
            lint_ctx.warn(
                f"Tool version is pre/suffixed by whitespace, this may cause errors: [{version}].",
                linter=cls.name(),
                node=tool_node,
            )


class ToolVersionValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml, tool_node = _tool_xml_and_root(tool_source)
        version = tool_source.parse_version() or ""
        parsed_version = parse_version(version)
        if version and not isinstance(parsed_version, LegacyVersion) and version == version.strip():
            lint_ctx.valid(f"Tool defines a version [{version}].", linter=cls.name(), node=tool_node)


class ToolNameMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        name = tool_source.parse_name()
        if not name:
            lint_ctx.error("Tool name is missing or empty.", linter=cls.name(), node=tool_node)


class ToolNameWhitespace(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        name = tool_source.parse_name()
        if name and name != name.strip():
            lint_ctx.warn(
                f"Tool name is pre/suffixed by whitespace, this may cause errors: [{name}].",
                linter=cls.name(),
                node=tool_node,
            )


class ToolNameValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        name = tool_source.parse_name()
        if name and name == name.strip():
            lint_ctx.valid(f"Tool defines a name [{name}].", linter=cls.name(), node=tool_node)


class ToolIDMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        tool_id = tool_source.parse_id()
        if not tool_id:
            lint_ctx.error("Tool does not define an id attribute.", linter=cls.name(), node=tool_node)


class ToolIDWhitespace(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        tool_id = tool_source.parse_id()
        if tool_id and re.search(r"\s", tool_id):
            lint_ctx.warn(
                f"Tool ID contains whitespace - this is discouraged: [{tool_id}].", linter=cls.name(), node=tool_node
            )


class ToolIDValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        tool_id = tool_source.parse_id()
        if tool_id and not re.search(r"\s", tool_id):
            lint_ctx.valid(f"Tool defines an id [{tool_id}].", linter=cls.name(), node=tool_node)


class ToolProfileInvalid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        profile = tool_source.parse_profile()
        profile_valid = PROFILE_PATTERN.match(profile) is not None
        if not profile_valid:
            lint_ctx.error(f"Tool specifies an invalid profile version [{profile}].", linter=cls.name(), node=tool_node)


class ToolProfileLegacy(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        profile = tool_source.parse_profile()
        profile_valid = PROFILE_PATTERN.match(profile) is not None
        if profile_valid and Version(profile) == Version("16.01"):
            lint_ctx.valid("Tool targets 16.01 Galaxy profile.", linter=cls.name(), node=tool_node)


class ToolProfileValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        profile = tool_source.parse_profile()
        profile_valid = PROFILE_PATTERN.match(profile) is not None
        if profile_valid and Version(profile) != Version("16.01"):
            lint_ctx.valid(f"Tool specifies profile version [{profile}].", linter=cls.name(), node=tool_node)


class RequirementNameMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        requirements, *_ = tool_source.parse_requirements()
        for r in requirements:
            if r.type != "package":
                continue
            if not r.name:
                lint_ctx.error("Requirement without name found", linter=cls.name(), node=tool_node)


class RequirementVersionMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        requirements, *_ = tool_source.parse_requirements()
        for r in requirements:
            if r.type != "package":
                continue
            if not r.version:
                lint_ctx.warn(f"Requirement {r.name} defines no version", linter=cls.name(), node=tool_node)


class RequirementVersionWhitespace(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        requirements, *_ = tool_source.parse_requirements()
        for r in requirements:
            if r.type != "package":
                continue
            if r.version and r.version != r.version.strip():
                lint_ctx.warn(
                    f"Requirement version contains whitespace, this may cause errors: [{r.version}].",
                    linter=cls.name(),
                    node=tool_node,
                )


class ResourceRequirementExpression(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        requirements, containers, resource_requirements, *_ = tool_source.parse_requirements()
        for rr in resource_requirements:
            if rr.runtime_required:
                lint_ctx.warn(
                    "Expressions in resource requirement not supported yet", linter=cls.name(), node=tool_node
                )


class BioToolsValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        xrefs = tool_source.parse_xrefs()
        for xref in xrefs:
            if xref["type"] != "bio.tools":
                continue
            metadata_source = ApiBiotoolsMetadataSource()
            if not metadata_source.get_biotools_metadata(xref["value"]):
                lint_ctx.warn(f'No entry {xref["value"]} in bio.tools.', linter=cls.name(), node=tool_node)


class EDAMTermsValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, tool_node = _tool_xml_and_root(tool_source)
        edam = load_edam_tree(None, "operation_", "topic_")
        terms = tool_source.parse_edam_operations() + tool_source.parse_edam_topics()
        for term in terms:
            if term not in edam:
                lint_ctx.warn(f"No entry '{term}' in EDAM.", linter=cls.name(), node=tool_node)



#!/usr/bin/env python

# . .venv/bin/activate; PYTHONPATH=lib python lib/galaxy/tool_util/client/landing.py -g http://localhost:8081 simple_workflow

import argparse
import random
import string
import sys
from dataclasses import dataclass
from typing import Optional

import requests
import yaml

from galaxy.util.resources import resource_string

DESCRIPTION = """
A small utility for demoing creation of tool and workflow landing endpoints.

This allows an external developer to create tool and workflow forms with
pre-populated parameters and handing them off with URLs to their clients/users.
"""
RANDOM_SECRET_LENGTH = 10


def load_default_catalog():
    name = __name__
    if __name__ == "__main__":
        name = "galaxy.tool_util.client.landing"
    catalog_yaml = resource_string(name, "landing_library.catalog.yml")
    return yaml.safe_load(catalog_yaml)


@dataclass
class Request:
    template_id: str
    catalog: str
    public: bool
    client_secret: Optional[str]
    galaxy_url: str


@dataclass
class Response:
    landing_url: str


def generate_claim_url(request: Request) -> Response:
    template_id = request.template_id
    catalog_path = request.catalog
    galaxy_url = request.galaxy_url
    client_secret = request.client_secret
    if client_secret == "__GEN__":
        client_secret = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(RANDOM_SECRET_LENGTH)
        )
    if catalog_path:
        with open(catalog_path) as f:
            catalog = yaml.safe_load(f)
    else:
        catalog = load_default_catalog()
    template = catalog[template_id]
    if "tool_id" in template:
        template_type = "tool"
    elif "workflow_id" in template:
        template_type = "workflow"
    elif isinstance(template["request_state"], list):
        template_type = "file"
    else:
        template_type = "data"
    if client_secret:
        template["client_secret"] = client_secret
    if request.public:
        template["public"] = True

    landing_request_url = f"{galaxy_url}/api/{template_type}_landings"
    raw_response = requests.post(
        landing_request_url,
        json=template,
    )
    try:
        raw_response.raise_for_status()
    except Exception:
        raise Exception("Request failed: %s", raw_response.json())
    response = raw_response.json()
    response_type = "workflow" if template_type == "workflow" else "tool"
    url = f"{galaxy_url}/{response_type}_landings/{response['uuid']}"
    if client_secret:
        url = url + f"?secret={client_secret}"
    return Response(url)


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("template_id")
    parser.add_argument(
        "-g",
        "--galaxy-url",
        dest="galaxy_url",
        default="https://usegalaxy.org/",
        help="Galxy target for the landing request",
    )
    parser.add_argument(
        "-c",
        "--catalog",
        dest="catalog",
        default=None,
        help="YAML catalog to load landing request templates from.",
    )
    parser.add_argument(
        "-s",
        "--secret",
        dest="secret",
        default=None,
        help="An optional client secret to verify the request against, set to __GEN__ to generate one at random for this request.",
    )
    parser.add_argument(
        "-p",
        "--public",
        dest="public",
        default=False,
        action="store_true",
        help="Declare the landing as public.",
    )
    return parser


def main(argv=None) -> None:
    if argv is None:
        argv = sys.argv[1:]

    args = arg_parser().parse_args(argv)
    request = Request(
        args.template_id,
        args.catalog,
        args.public,
        args.secret,
        args.galaxy_url,
    )
    response = generate_claim_url(request)
    print(f"Your customized form is located at {response.landing_url}")


if __name__ == "__main__":
    main()



from ._types import (
    Annotated,
    AssertionParameter,
    Delta,
    Max,
    Min,
    Negate,
    NEGATE_DEFAULT,
    OptionalXmlInt,
    OutputBytes,
)
from ._util import _assert_number


def assert_has_size(
    output_bytes: OutputBytes,
    value: Annotated[OptionalXmlInt, AssertionParameter("Deprecated alias for `size`", xml_type="Bytes")] = None,
    size: Annotated[
        OptionalXmlInt,
        AssertionParameter(
            "Desired size of the output (in bytes), can be suffixed by ``(k|M|G|T|P|E)i?``", xml_type="Bytes"
        ),
    ] = None,
    delta: Delta = 0,
    min: Min = None,
    max: Max = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """Asserts the specified output has a size of the specified value

    Attributes size and value or synonyms though value is considered deprecated.
    The size optionally allows for absolute (``delta``) difference.
    """
    output_size = len(output_bytes)
    if size is None:
        size = value
    _assert_number(
        output_size,
        size,
        delta,
        min,
        max,
        negate,
        "{expected} file size of {n}+-{delta}",
        "{expected} file size to be in [{min}:{max}]",
    )



#!/usr/bin/env python
"""Build a mulled images for a tool source (Galaxy or CWL tool).

Examples:

Build mulled images for requirements defined in a tool:

    mulled-build-tool build path/to/tool_file.xml

"""

from typing import (
    List,
    TYPE_CHECKING,
)

from galaxy.tool_util.parser import get_tool_source
from ._cli import arg_parser
from .mulled_build import (
    add_build_arguments,
    add_single_image_arguments,
    args_to_mull_targets_kwds,
    mull_targets,
    target_str_to_targets,
)

if TYPE_CHECKING:
    from galaxy.tool_util.deps.conda_util import CondaTarget
    from galaxy.util.path import StrPath


def _mulled_build_tool(tool: "StrPath", args):
    tool_source = get_tool_source(tool)
    requirements, *_ = tool_source.parse_requirements()
    targets = requirements_to_mulled_targets(requirements)
    kwds = args_to_mull_targets_kwds(args)
    mull_targets(targets, **kwds)


def main(argv=None) -> None:
    """Main entry-point for the CLI tool."""
    parser = arg_parser(argv, globals())
    add_build_arguments(parser)
    add_single_image_arguments(parser)
    parser.add_argument("command", metavar="COMMAND", help="Command (build-and-test, build, all)")
    parser.add_argument("tool", metavar="TOOL", default=None, help="Path to tool to build mulled image for.")
    args = parser.parse_args()
    _mulled_build_tool(args.tool, args)


def requirements_to_mulled_targets(requirements) -> List["CondaTarget"]:
    """Convert Galaxy's representation of requirements into a list of CondaTarget objects.

    Only package requirements are retained.
    """
    package_requirements = [r for r in requirements if r.type == "package"]
    target_str = ",".join(f"{r.name}={r.version}" if r.version else r.name for r in package_requirements)
    targets = target_str_to_targets(target_str)
    return targets


__all__ = ("main", "requirements_to_mulled_targets")


if __name__ == "__main__":
    main()



from typing import (
    Type,
    TypeVar,
)

from galaxy.tool_util_models import ParsedTool
from .parameters import input_models_for_tool_source
from .parser.interface import (
    ToolSource,
)
from .parser.output_objects import from_tool_source


def parse_tool(tool_source: ToolSource) -> ParsedTool:
    return parse_tool_custom(tool_source, ParsedTool)


P = TypeVar("P", bound=ParsedTool)


def parse_tool_custom(tool_source: ToolSource, model_type: Type[P]) -> P:
    id = tool_source.parse_id()
    version = tool_source.parse_version()
    name = tool_source.parse_name()
    description = tool_source.parse_description()
    inputs = input_models_for_tool_source(tool_source).parameters
    outputs = from_tool_source(tool_source)
    citations = tool_source.parse_citations()
    license = tool_source.parse_license()
    profile = tool_source.parse_profile()
    edam_operations = tool_source.parse_edam_operations()
    edam_topics = tool_source.parse_edam_topics()
    xrefs = tool_source.parse_xrefs()
    help = tool_source.parse_help()

    return model_type(
        id=id,
        version=version,
        name=name,
        description=description,
        profile=profile,
        inputs=inputs,
        outputs=outputs,
        license=license,
        citations=citations,
        edam_operations=edam_operations,
        edam_topics=edam_topics,
        xrefs=xrefs,
        help=help,
    )



import json
from typing import (
    Any,
    Callable,
    cast,
)

from ._types import (
    Annotated,
    AssertionParameter,
    Output,
)

PropertyVisitor = Callable[[str, Any], Any]

Property = Annotated[str, AssertionParameter("The property name to search the JSON document for.")]
Text = Annotated[str, AssertionParameter("The expected text value of the target JSON attribute.")]
Value = Annotated[
    str, AssertionParameter("The expected JSON value of the target JSON attribute (as a JSON encoded string).")
]


def any_in_tree(f: PropertyVisitor, json_tree: Any):
    if isinstance(json_tree, list):
        for element in json_tree:
            if any_in_tree(f, element):
                return True

    elif isinstance(json_tree, dict):
        for key, value in json_tree.items():
            if f(cast(str, key), value):
                return True
            elif any_in_tree(f, value):
                return True

    return False


def assert_has_json_property_with_value(
    output: Output,
    property: Property,
    value: Value,
):
    """Asserts the JSON document contains a property or key with the specified JSON value.

    ```xml
    <has_json_property_with_value property="skipped_columns" value="[1, 3, 5]" />
    ```
    """
    output_json = _assert_json_and_load(output)
    expected_value = _assert_json_and_load(value)

    def is_property(key, value):
        return key == property and value == expected_value

    assert any_in_tree(is_property, output_json), f"Failed to find property [{property}] with JSON value [{value}]"


def assert_has_json_property_with_text(
    output: Output,
    property: Property,
    text: Text,
):
    """Asserts the JSON document contains a property or key with the specified text (i.e. string) value.

    ```xml
    <has_json_property_with_text property="color" text="red" />
    ```
    """
    output_json = _assert_json_and_load(output)

    def is_property(key, value):
        return key == property and value == text

    assert any_in_tree(is_property, output_json), f"Failed to find property [{property}] with text [{text}]"


def _assert_json_and_load(json_str: str):
    try:
        return json.loads(json_str)
    except Exception:
        raise AssertionError(f"Failed to parse JSON from {json_str[0:1024]}.")



from string import Template

SIMPLE_TOOL_WITH_MACRO = """<tool id="tool_with_macro" name="macro_annotation" version="@WRAPPER_VERSION@">
    <expand macro="inputs" />
    <macros>
        <import>external.xml</import>
    </macros>
</tool>"""

SIMPLE_MACRO = Template("""
<macros>
    <token name="@WRAPPER_VERSION@">$tool_version</token>
    <macro name="inputs">
        <inputs/>
    </macro>
</macros>
""")

VALID_XML_VALIDATORS = [
    """<validator type="empty_dataset" />""",
    """<validator type="empty_dataset" message="foobar" />""",
    """<validator type="empty_dataset" message="foobar" negate="true" />""",
    """<validator type="expression">value == 7</validator>""",
    """<validator type="regex">mycoolexpression</validator>""",
    """<validator type="in_range" min="3.1" max="3.8" />""",
    """<validator type="in_range" min="3.1" />""",
    """<validator type="in_range" min="3.1" max="3.8" exclude_min="false" exclude_max="true" />""",
    """<validator type="length" min="3" />""",
    """<validator type="length" max="7" />""",
    """<validator type="length" min="2" max="7" negate="true" />""",
    """<validator type="empty_field" />""",
    """<validator type="empty_extra_files_path" />""",
    """<validator type="no_options" />""",
    """<validator type="unspecified_build" />""",
    """<validator type="dataset_ok_validator" />""",
    """<validator type="metadata" check="foo, bar" />""",
    """<validator type="metadata" skip="name, dbkey" />""",
    """<validator type="dataset_metadata_equal" metadata_name="foobar" value="moocow" />""",
    """<validator type="dataset_metadata_equal" metadata_name="foobar" value_json="null" />""",
    """<validator type="dataset_metadata_in_range" metadata_name="foobar" min="4.5" max="7.8" />""",
    """<validator type="dataset_metadata_in_range" metadata_name="foobar" min="4.5" max="7.8" include_min="true" />""",
    """<validator type="dataset_metadata_in_data_table" metadata_name="foobar" metadata_column="3" table_name="mycooltable" />""",
    """<validator type="dataset_metadata_not_in_data_table" metadata_name="foobar" metadata_column="3" table_name="mycooltable" />""",
    """<validator type="value_in_data_table" metadata_column="3" table_name="mycooltable" />""",
    """<validator type="value_in_data_table" table_name="mycooltable" />""",
    """<validator type="value_not_in_data_table" metadata_column="3" table_name="mycooltable" />""",
    """<validator type="value_not_in_data_table" table_name="mycooltable" />""",
]

INVALID_XML_VALIDATORS = [
    """<validator type="unknown" />""",
    """<validator type="empty_datasetx" />""",
    """<validator type="expression" />""",
    """<validator type="empty_dataset" message="foobar" negate="NOTABOOLVALUE" />""",
    """<validator type="regex" />""",
    """<validator type="in_range" min="3.1" max="3.8" exclude_min="false" exclude_max="foobar" />""",
    """<validator type="in_range" min="notanumber" max="3.8" />""",
    """<validator type="length" min="notanumber" />""",
    """<validator type="length" max="notanumber" />""",
    """<validator type="length" min="2" max="7" negate="notabool" />""",
    """<validator type="dataset_metadata_equal" />""",
    """<validator type="dataset_metadata_equal" metadaata_name="foobar" />""",
    """<validator type="dataset_metadata_equal" metadaata_name="foobar" value_json="undefined" />""",
    """<validator type="dataset_metadata_in_range" metadata_name="foobar" min="4.5" max="7.8" include_min="notabool" />"""
    """<validator type="dataset_metadata_in_range"  min="4.5" max="7.8" />"""
    """<validator type="dataset_metadata_in_data_table" metadata_name="foobar" metadata_column="3" />""",
    """<validator type="dataset_metadata_in_data_table" metadata_column="3" table_name="mycooltable" />""",
    """<validator type="dataset_metadata_not_in_data_table" metadata_name="foobar" metadata_column="3" />""",
    """<validator type="dataset_metadata_not_in_data_table" metadata_column="3" table_name="mycooltable" />""",
]



"""This module contains a linting functions for tool tests."""

from io import StringIO
from typing import (
    Iterator,
    List,
    Tuple,
    TYPE_CHECKING,
)

from packaging.version import Version

from galaxy.tool_util.lint import Linter
from galaxy.tool_util.parameters import validate_test_cases_for_tool_source
from galaxy.tool_util.verify.parse import tag_structure_to_that_structure
from galaxy.tool_util_models.assertions import (
    assertion_list,
    relaxed_assertion_list,
)
from galaxy.util import asbool
from ._util import is_datasource

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser.interface import ToolSource
    from galaxy.util.etree import Element

lint_tool_types = ["default", "data_source", "manage_data"]


class TestsMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        root = tool_xml.find("./tests")
        if root is None:
            root = tool_xml.getroot()
        if len(tests) == 0 and not is_datasource(tool_xml):
            lint_ctx.warn("No tests found, most tools should define test cases.", linter=cls.name(), node=root)


class TestsMissingDatasource(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        root = tool_xml.find("./tests")
        if root is None:
            root = tool_xml.getroot()
        if len(tests) == 0 and is_datasource(tool_xml):
            lint_ctx.info("No tests found, that should be OK for data_sources.", linter=cls.name(), node=root)


class TestsAssertsMultiple(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            # TODO same would be nice also for assert_contents
            for ta in ("assert_stdout", "assert_stderr", "assert_command"):
                if len(test.findall(ta)) > 1:
                    lint_ctx.error(
                        f"Test {test_idx}: More than one {ta} found. Only the first is considered.",
                        linter=cls.name(),
                        node=test,
                    )


class TestsAssertsHasNQuant(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for a in test.xpath(
                ".//*[self::assert_contents or self::assert_stdout or self::assert_stderr or self::assert_command]//*"
            ):
                if a.tag not in ["has_n_lines", "has_n_columns"]:
                    continue
                if not (set(a.attrib) & {"n", "min", "max"}):
                    lint_ctx.error(
                        f"Test {test_idx}: '{a.tag}' needs to specify 'n', 'min', or 'max'", linter=cls.name(), node=a
                    )


class TestsAssertsHasSizeQuant(Linter):
    """
    has_size needs at least one of size (or value), min, max
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for a in test.xpath(
                ".//*[self::assert_contents or self::assert_stdout or self::assert_stderr or self::assert_command]//*"
            ):
                if a.tag != "has_size":
                    continue
                if len(set(a.attrib) & {"value", "size", "min", "max"}) == 0:
                    lint_ctx.error(
                        f"Test {test_idx}: '{a.tag}' needs to specify 'size', 'min', or 'max'",
                        linter=cls.name(),
                        node=a,
                    )


class TestsAssertsHasSizeOrValueQuant(Linter):
    """
    has_size should not have size and value
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for a in test.xpath(
                ".//*[self::assert_contents or self::assert_stdout or self::assert_stderr or self::assert_command]//*"
            ):
                if a.tag != "has_size":
                    continue
                if "value" in a.attrib and "size" in a.attrib:
                    lint_ctx.error(
                        f"Test {test_idx}: '{a.tag}' must not specify 'value' and 'size'",
                        linter=cls.name(),
                        node=a,
                    )


class TestsAssertionValidation(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        try:
            raw_tests_dict = tool_source.parse_tests_to_dict()
        except Exception:
            lint_ctx.warn("Failed to parse test dictionaries from tool - cannot lint assertions")
            return
        assert "tests" in raw_tests_dict
        # This really only allows coercion from strings, the values themselves will still be validated
        assert_list_model = relaxed_assertion_list if tool_source.language == "xml" else assertion_list
        for test_idx, test in enumerate(raw_tests_dict["tests"], start=1):
            # TODO: validate command, command_version, element tests. What about children?
            for output in test["outputs"]:
                asserts_raw = output.get("attributes", {}).get("assert_list") or []
                processed_assertions = []
                for raw_assert in asserts_raw:
                    processed_assertions.append(tag_structure_to_that_structure(raw_assert))
                try:
                    assert_list_model.model_validate(processed_assertions)
                except Exception as e:
                    error_str = _cleanup_pydantic_error(e)
                    lint_ctx.warn(
                        f"Test {test_idx}: failed to validate assertions. Validation errors are [{error_str}]",
                        linter=cls.name(),
                    )


class TestsCaseValidation(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        profile = tool_source.parse_profile()
        lint_log = lint_ctx.warn if Version(profile) < Version("24.2") else lint_ctx.error

        try:
            validation_results = validate_test_cases_for_tool_source(tool_source, use_latest_profile=True)
        except Exception as e:
            lint_log(
                f"Serious problem parsing tool source or tests - cannot validate test cases. The exception is [{e}]",
                linter=cls.name(),
            )
            return
        for test_idx, validation_result in enumerate(validation_results, start=1):
            error = validation_result.validation_error
            if error:
                error_str = _cleanup_pydantic_error(error)
                lint_log(
                    f"Test {test_idx}: failed to validate test parameters against inputs - tests won't run on a modern Galaxy tool profile version. Validation errors are [{error_str}]",
                    linter=cls.name(),
                )


def _cleanup_pydantic_error(error) -> str:
    full_validation_error = f"{error}"
    new_error = StringIO("")
    for line in full_validation_error.splitlines():
        # this repeated over and over isn't useful in the context of how we're building the dynamic models,
        # tool authors should not be looking up pydantic docs on models they cannot even really inspect
        if line.strip().startswith("For further information visit https://errors.pydantic"):
            continue
        else:
            new_error.write(f"{line}\n")
    return new_error.getvalue().strip()


class TestsExpectNumOutputs(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            # check if expect_num_outputs is set if there are outputs with filters
            # (except for tests with expect_failure .. which can't have test outputs)
            has_no_filter = (
                tool_xml.find("./outputs/data/filter") is None and tool_xml.find("./outputs/collection/filter") is None
            )
            if not (
                has_no_filter or "expect_num_outputs" in test.attrib or asbool(test.attrib.get("expect_failure", False))
            ):
                lint_ctx.warn(
                    f"Test {test_idx}: should specify 'expect_num_outputs' if outputs have filters",
                    linter=cls.name(),
                    node=test,
                )


class TestsParamInInputs(Linter):
    """
    really simple linter that test parameters are also present in the inputs
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for param in test.findall("param"):
                name = param.attrib.get("name", None)
                if not name:
                    continue
                name = name.split("|")[-1]
                xpaths = [f"@name='{name}'", f"@argument='{name}'", f"@argument='-{name}'", f"@argument='--{name}'"]
                if "_" in name:
                    xpaths += [
                        f"@argument='{name.replace('_', '-')}'",
                        f"@argument='-{name.replace('_', '-')}'",
                        f"@argument='--{name.replace('_', '-')}'",
                    ]
                found = False
                for xp in xpaths:
                    inxpath = f".//inputs//param[{xp}]"
                    inparam = tool_xml.findall(inxpath)
                    if len(inparam) > 0:
                        found = True
                        break
                if not found:
                    lint_ctx.error(
                        f"Test {test_idx}: Test param {name} not found in the inputs", linter=cls.name(), node=param
                    )


class TestsOutputName(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            # note output_collections are covered by xsd, but output is not required to have one by xsd
            for output in test.findall("output"):
                if not output.attrib.get("name", None):
                    lint_ctx.error(
                        f"Test {test_idx}: Found {output.tag} tag without a name defined.",
                        linter=cls.name(),
                        node=output,
                    )


class TestsOutputDefined(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        output_data_or_collection = _collect_output_names(tool_xml)
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for output in test.findall("output") + test.findall("output_collection"):
                name = output.attrib.get("name", None)
                if not name:
                    continue
                if name not in output_data_or_collection:
                    lint_ctx.error(
                        f"Test {test_idx}: Found {output.tag} tag with unknown name [{name}], valid names {list(output_data_or_collection)}",
                        linter=cls.name(),
                        node=output,
                    )


class TestsOutputCorresponding(Linter):
    """
    Linter checking if test/output corresponds to outputs/data
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        output_data_or_collection = _collect_output_names(tool_xml)
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for output in test.findall("output") + test.findall("output_collection"):
                name = output.attrib.get("name", None)
                if not name:
                    continue
                if name not in output_data_or_collection:
                    continue

                # - test/collection to outputs/output_collection
                corresponding_output = output_data_or_collection[name]
                if output.tag == "output" and corresponding_output.tag != "data":
                    lint_ctx.error(
                        f"Test {test_idx}: test output {name} does not correspond to a 'data' output, but a '{corresponding_output.tag}'",
                        linter=cls.name(),
                        node=output,
                    )


class TestsOutputCollectionCorresponding(Linter):
    """
    Linter checking if test/collection corresponds to outputs/output_collection
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        output_data_or_collection = _collect_output_names(tool_xml)
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for output in test.findall("output") + test.findall("output_collection"):
                name = output.attrib.get("name", None)
                if not name:
                    continue
                if name not in output_data_or_collection:
                    continue

                # - test/collection to outputs/output_collection
                corresponding_output = output_data_or_collection[name]
                if output.tag == "output_collection" and corresponding_output.tag != "collection":
                    lint_ctx.error(
                        f"Test {test_idx}: test collection output '{name}' does not correspond to a 'output_collection' output, but a '{corresponding_output.tag}'",
                        linter=cls.name(),
                        node=output,
                    )


class TestsOutputCompareAttrib(Linter):
    """
    Linter checking compatibility of output attributes with the value
    of the compare attribute
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        COMPARE_COMPATIBILITY = {
            "sort": ["diff", "re_match", "re_match_multiline"],
            "lines_diff": ["diff", "re_match", "contains"],
            "decompress": ["diff"],
            "delta": ["sim_size"],
            "delta_frac": ["sim_size"],
            "metric": ["image_diff"],
            "eps": ["image_diff"],
        }
        for test_idx, test in enumerate(tests, start=1):
            for output in test.xpath(".//*[self::output or self::element or self::discovered_dataset]"):
                compare = output.get("compare", "diff")
                for attrib in COMPARE_COMPATIBILITY:
                    if attrib in output.attrib and compare not in COMPARE_COMPATIBILITY[attrib]:
                        lint_ctx.error(
                            f'Test {test_idx}: Attribute {attrib} is incompatible with compare="{compare}".',
                            linter=cls.name(),
                            node=output,
                        )


class TestsOutputCheckDiscovered(Linter):
    """
    Linter checking that discovered elements of outputs are tested with
    a count attribute or listing some discovered_dataset
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        output_data_or_collection = _collect_output_names(tool_xml)
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for output in test.findall("output"):
                name = output.attrib.get("name", None)
                if not name:
                    continue
                if name not in output_data_or_collection:
                    continue

                # - test/collection to outputs/output_collection
                corresponding_output = output_data_or_collection[name]
                discover_datasets = corresponding_output.find(".//discover_datasets")
                if discover_datasets is None:
                    continue
                if (
                    "count" not in output.attrib
                    and "min" not in output.attrib
                    and "max" not in output.attrib
                    and output.find("./discovered_dataset") is None
                ):
                    lint_ctx.error(
                        f"Test {test_idx}: test output '{name}' must have a 'count/min/max' attribute and/or 'discovered_dataset' children",
                        linter=cls.name(),
                        node=output,
                    )


class TestsOutputCollectionCheckDiscovered(Linter):
    """
    Linter checking that discovered elements of output collections
    are tested with a count attribute or listing some elements
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        output_data_or_collection = _collect_output_names(tool_xml)
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for output in test.findall("output_collection"):
                name = output.attrib.get("name", None)
                if not name:
                    continue
                if name not in output_data_or_collection:
                    continue
                # - test/collection to outputs/output_collection
                corresponding_output = output_data_or_collection[name]
                if corresponding_output.find(".//discover_datasets") is None:
                    continue
                if (
                    "count" not in output.attrib
                    and "min" not in output.attrib
                    and "max" not in output.attrib
                    and output.find("./element") is None
                ):
                    lint_ctx.error(
                        f"Test {test_idx}: test collection '{name}' must have a 'count/min/max' attribute or 'element' children",
                        linter=cls.name(),
                        node=output,
                    )


class TestsOutputCollectionCheckDiscoveredNested(Linter):
    """ """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        output_data_or_collection = _collect_output_names(tool_xml)
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            for output in test.findall("output_collection"):
                name = output.attrib.get("name", None)
                if not name:
                    continue
                if name not in output_data_or_collection:
                    continue

                # - test/collection to outputs/output_collection
                corresponding_output = output_data_or_collection[name]
                if corresponding_output.find(".//discover_datasets") is None:
                    continue
                if corresponding_output.get("type", "") in ["list:list", "list:paired"]:
                    nested_elements = output.find("./element/element")
                    elements_with_count = output.xpath("./element[@count or @min or @max]")
                    if nested_elements is None and not elements_with_count:
                        lint_ctx.error(
                            f"Test {test_idx}: test collection '{name}' must contain nested 'element' tags and/or element children with a 'count/min/max' attribute",
                            linter=cls.name(),
                            node=output,
                        )


class TestsOutputFailing(Linter):
    """ """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            if not asbool(test.attrib.get("expect_failure", False)):
                continue
            if test.find("output") is not None or test.find("output_collection") is not None:
                lint_ctx.error(
                    f"Test {test_idx}: Cannot specify outputs in a test expecting failure.",
                    linter=cls.name(),
                    node=test,
                )


class TestsExpectNumOutputsFailing(Linter):
    """ """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in enumerate(tests, start=1):
            if not asbool(test.attrib.get("expect_failure", False)):
                continue
            if test.find("output") is not None or test.find("output_collection") is not None:
                continue
            if "expect_num_outputs" in test.attrib:
                lint_ctx.error(
                    f"Test {test_idx}: Cannot make assumptions on the number of outputs in a test expecting failure.",
                    linter=cls.name(),
                    node=test,
                )


class TestsHasExpectations(Linter):
    """ """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tests = tool_xml.findall("./tests/test")
        for test_idx, test in _iter_tests(tests, valid=False):
            lint_ctx.warn(
                f"Test {test_idx}: No outputs or expectations defined for tests, this test is likely invalid.",
                linter=cls.name(),
                node=test,
            )


class TestsNoValid(Linter):
    """ """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        general_node = tool_xml.find("./tests")
        if general_node is None:
            general_node = tool_xml.getroot()
        tests = tool_xml.findall("./tests/test")
        if not tests:
            return
        num_valid_tests = len(list(_iter_tests(tests, valid=True)))
        if num_valid_tests or is_datasource(tool_xml):
            lint_ctx.valid(f"{num_valid_tests} test(s) found.", linter=cls.name(), node=general_node)


class TestsValid(Linter):
    """ """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        general_node = tool_xml.find("./tests")
        if general_node is None:
            general_node = tool_xml.getroot()
        tests = tool_xml.findall("./tests/test")
        if not tests:
            return
        num_valid_tests = len(list(_iter_tests(tests, valid=True)))
        if not (num_valid_tests or is_datasource(tool_xml)):
            lint_ctx.warn("No valid test(s) found.", linter=cls.name(), node=general_node)


def _iter_tests(tests: List["Element"], valid: bool) -> Iterator[Tuple[int, "Element"]]:
    for test_idx, test in enumerate(tests, start=1):
        is_valid = False
        is_valid |= bool(set(test.attrib) & {"expect_failure", "expect_exit_code", "expect_num_outputs"})
        for ta in ("assert_stdout", "assert_stderr", "assert_command"):
            if test.find(ta) is not None:
                is_valid = True
        found_output_test = test.find("output") is not None or test.find("output_collection") is not None
        if asbool(test.attrib.get("expect_failure", False)):
            if found_output_test or "expect_num_outputs" in test.attrib:
                continue
        is_valid |= found_output_test
        if is_valid == valid:
            yield (test_idx, test)


def _collect_output_names(tool_xml):
    """
    determine dict mapping the names of data and collection outputs to the
    corresponding nodes
    """
    output_data_or_collection = {}
    outputs = tool_xml.findall("./outputs")
    if len(outputs) == 1:
        for output in list(outputs[0]):
            name = output.attrib.get("name", None)
            if not name:
                continue
            output_data_or_collection[name] = output
    return output_data_or_collection
